"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


Module with classes and methods to visualized structural geology data and potential fields of the regional modelling based on
the potential field method.
Tested on Ubuntu 14

Created on 23/09/2016

@author: Miguel de la Varga, Elisa Heim
"""

import warnings
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.colors import LightSource
from matplotlib.ticker import FixedFormatter, FixedLocator
import seaborn as sns
from os import path
import sys

# This is for sphenix to find the packages
sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
from gempy_legacy.core.solution import Solution
from  gempy_legacy.plot import helpers as plothelp
sns.set_context('talk')
plt.style.use(['seaborn-v0_8-white', 'seaborn-v0_8-talk'])
from scipy.interpolate import RegularGridInterpolator
import matplotlib.patches as mpatches



class PlotData2D:
    def __init__(self, model):
        self.model = model
        self._color_lot = dict(zip(self.model._surfaces.df['surface'], self.model._surfaces.df['color']))
        self._cmap = mcolors.ListedColormap(list(self.model._surfaces.df['color']))
        self._norm = mcolors.Normalize(vmin=0.5, vmax=len(self._cmap.colors) + 0.5)

    def get_plot_data(self, series="all", at="everywhere", direction=None, cell_number=None, radius='default',
                      show_all_data=False):

        if radius == 'default':
            radius = None
        else:
            if not isinstance(radius, int):
                raise AttributeError('You need to pass a number (in model extent) for the radius to take more '
                                     'or less data into account.')

        if series == "all":
            series_to_plot_i = self.model._surface_points.df[self.model._surface_points.df["series"].
                isin(self.model._stack.df.index.values)]
            series_to_plot_f = self.model._orientations.df[self.model._orientations.df["series"].
                isin(self.model._stack.df.index.values)]

        else:
            series_to_plot_i = self.model._surface_points.df[self.model._surface_points.df["series"] == series]
            series_to_plot_f = self.model._orientations.df[self.model._orientations.df["series"] == series]

        if show_all_data:
            at = 'everywhere'
        if type(at) == str:
            if at == 'topography':
                mask_surfpoints, mask_orient = self.get_mask_surface_data(radius=radius)
            elif at == 'block_section':
                mask_surfpoints, mask_orient = self.get_mask_block_section(cell_number=cell_number, direction=direction,
                                                                           radius=radius)
            elif at == 'everywhere':
                mask_surfpoints = np.ones(series_to_plot_i.shape[0], dtype=bool)
                mask_orient = np.ones(series_to_plot_f.shape[0], dtype=bool)
            else:  # see if it is a section name
                try:
                    j = np.where(self.model._grid.sections.names == at)[0][0]
                    mask_surfpoints, mask_orient = self.get_mask_sections(j, radius=radius)

                except:
                    raise AttributeError  # 'must be topography, a section name or block_section'

        elif type(at) == list:  # should be a list of section names but must be asserted
            try:
                mask_surfpoints = np.zeros(series_to_plot_i.shape[0], dtype=bool)
                mask_orient = np.zeros(series_to_plot_f.shape[0], dtype=bool)
                for i in at:
                    j = np.where(self.model._grid.sections.names == i)[0][0]
                    mask_surfpoints_i, mask_orient_i = self.get_mask_sections(j, radius=radius)
                    mask_surfpoints[mask_surfpoints_i] = True
                    mask_orient[mask_orient_i] = True
            except AttributeError:
                print('must be topography, a section name or block_section')
        else:
            print('problem')

        return series_to_plot_i[mask_surfpoints], series_to_plot_f[mask_orient]

    def plot_data(self, cell_number=2, direction="y", data_type='all', series="all", show_legend=True, ve=1,
                  at='everywhere', radius='default', show_all_data=False, **kwargs):
        """
        Plot the projecton of the raw data (surface_points and orientations) in 2D following a
        specific directions

        Args:
            direction(str): xyz. Caartesian direction to be plotted
            data_type (str): type of data to plot. 'all', 'surface_points' or 'orientations'
            series(str): series to plot
            ve(float): Vertical exageration
            show_all_data:
            at:
            **kwargs: seaborn lmplot key arguments. (TODO: adding the link to them)

        Returns:
            Data plot

        """

        if 'scatter_kws' not in kwargs:
            kwargs['scatter_kws'] = {"marker": "o",
                                     "s": 100,
                                     "edgecolors": "black",
                                     "linewidths": 1}

        x, y, Gx, Gy = self._slice(direction)[4:]
        extent = self._slice(direction)[3]

        aspect = (extent[1] - extent[0]) / (extent[3] - extent[2])
        # apply vertical exageration
        if direction == 'x' or direction == 'y':
            aspect /= ve

        min_axis = 'width' if aspect < 1 else 'height'

        plot_surfpoints, plot_orient = self.get_plot_data(series=series, at=at, direction=direction,
                                                          radius=radius, cell_number=cell_number,
                                                          show_all_data=show_all_data)

        if data_type == 'all':
            self._plot_surface_points(x, y, plot_surfpoints, aspect, extent, kwargs)
            self._plot_orientations(x, y, Gx, Gy, plot_orient, min_axis, extent, False)

        if data_type == 'surface_points':
            self._plot_surface_points(x, y, plot_surfpoints, aspect, extent, kwargs)

        if data_type == 'orientations':
            self._plot_orientations(x, y, Gx, Gy, plot_orient, min_axis, extent, True, aspect)

        plt.xlim(extent[0], extent[1])
        plt.ylim(extent[2], extent[3])
        plt.xlabel(x)
        plt.ylabel(y)

    def plot_section_data(self, section_name, show_all_data=False, radius='default', **kwargs):
        if show_all_data:
            at = 'everywhere'
        else:
            at = section_name
        plot_surfpoints, plot_orient = self.get_plot_data(at=at, radius=radius)
        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        # convert the data to the new coord system
        # define x,y, Gx, Gy #orientations is difficult?
        # direction is to fix plot to x and z column, but we replace x column with projected values
        extent = [0, self.model._grid.sections.dist[j][0],
                  self.model._grid.regular_grid.extent[4], self.model._grid.regular_grid.extent[5]]

        a = self.model._grid.sections.points[j][0]  # startpoint = 0, we need distance to that point
        bs_i = plot_surfpoints[['X', 'Y']].values
        bs_o = plot_orient[['X', 'Y']].values
        new_x_points_i = np.linalg.norm(bs_i - a, axis=1)
        new_x_points_o = np.linalg.norm(bs_o - a, axis=1)
        plot_surfpoints['X'] = new_x_points_i
        plot_orient['X'] = new_x_points_o
        y = 'Z'
        x = 'X'
        aspect = (extent[1] - extent[0]) / (extent[3] - extent[2])
        kwargs['scatter_kws'] = {"marker": "o",
                                 "s": 100,
                                 "edgecolors": "black",
                                 "linewidths": 1}
        # print(plot_surfpoints)
        self._plot_surface_points(x, y, plot_surfpoints, aspect, extent, kwargs)
        Gx = 'G_x'
        Gy = 'G_z'
        min_axis = 'width' if aspect < 1 else 'height'
        self._plot_orientations(x, y, Gx, Gy, plot_orient, min_axis, extent, False, aspect)
        warnings.warn('the orientations are not converted to apparent dip.')

    def _plot_surface_points(self, x, y, series_to_plot_i, aspect, extent, kwargs):
        if series_to_plot_i.shape[0] != 0:
            # size = fig.get_size_inches() * fig.dpi
            # print(size)
            # print(aspect)
            try:
                p = sns.FacetGrid(series_to_plot_i, hue="surface",
                                  palette=self._color_lot,
                                  ylim=[extent[2], extent[3]],
                                  xlim=[extent[0], extent[1]],
                                  legend_out=False,
                                  aspect=aspect,
                                  height=6)
            except KeyError:  # for kriging dataframes
                p = sns.FacetGrid(series_to_plot_i, hue=None,
                                  palette='k',
                                  ylim=[extent[2], extent[3]],
                                  xlim=[extent[0], extent[1]],
                                  legend_out=False,
                                  aspect=aspect,
                                  height=6)

            p.map(plt.scatter, x, y,
                  **kwargs['scatter_kws'],
                  zorder=10)
        else:
            self._show_legend = True

    def _plot_orientations(self, x, y, Gx, Gy, series_to_plot_f, min_axis, extent, p, aspect=None, ax=None):
        if series_to_plot_f.shape[0] != 0:
            # print('hello')
            if p is False:
                # size = fig.get_size_inches() * fig.dpi
                # print('before plot orient', size)
                surflist = list(series_to_plot_f['surface'].unique())
                for surface in surflist:
                    to_plot = series_to_plot_f[series_to_plot_f['surface'] == surface]
                    plt.quiver(to_plot[x], to_plot[y],
                               to_plot[Gx], to_plot[Gy],
                               pivot="tail", scale_units=min_axis, scale=30, color=self._color_lot[surface],
                               edgecolor='k', headwidth=8, linewidths=1)
                    # ax.Axes.set_ylim([extent[2], extent[3]])
                    # ax.Axes.set_xlim([extent[0], extent[1]])
                # fig = plt.gcf()
                # fig.set_size_inches(20,10)
                # if aspect is not None:
                # ax = plt.gca()
                # ax.set_aspect(aspect)

            else:
                p = sns.FacetGrid(series_to_plot_f, hue="surface",
                                  palette=self._color_lot,
                                  ylim=[extent[2], extent[3]],
                                  xlim=[extent[0], extent[1]],
                                  legend_out=False,
                                  aspect=aspect,
                                  height=6)
                p.map(plt.quiver, x, y, Gx, Gy, pivot="tail", scale_units=min_axis, scale=10, edgecolor='k',
                      headwidth=4, linewidths=1)
        else:
            # print('no orient')
            pass

        # size = fig.get_size_inches() * fig.dpi
        # print('after plot_orientations', size)

    def _slice(self, direction, cell_number=25):
        """
        Slice the 3D array (blocks or scalar field) in the specific direction selected in the plot functions

        """
        _a, _b, _c = (slice(0, self.model._grid.regular_grid.resolution[0]),
                      slice(0, self.model._grid.regular_grid.resolution[1]),
                      slice(0, self.model._grid.regular_grid.resolution[2]))

        if direction == "x":
            _a, x, y, Gx, Gy = cell_number, "Y", "Z", "G_y", "G_z"
            extent_val = self.model._grid.regular_grid.extent[[2, 3, 4, 5]]
        elif direction == "y":
            _b, x, y, Gx, Gy = cell_number, "X", "Z", "G_x", "G_z"
            extent_val = self.model._grid.regular_grid.extent[[0, 1, 4, 5]]
        elif direction == "z":
            _c, x, y, Gx, Gy = cell_number, "X", "Y", "G_x", "G_y"
            extent_val = self.model._grid.regular_grid.extent[[0, 1, 2, 3]]
        else:
            raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")
        return _a, _b, _c, extent_val, x, y, Gx, Gy

    def _slice2D(self, cell_number, direction):
        if direction == 'x':
            _slice = np.s_[cell_number, :, :]
            extent = self.model._grid.regular_grid.extent[[2, 3, 4, 5]]
        elif direction == 'y':
            _slice = np.s_[:, cell_number, :]
            extent = self.model._grid.regular_grid.extent[[0, 1, 4, 5]]
        elif direction == 'z':
            _slice = np.s_[:, :, cell_number]
            extent = self.model._grid.regular_grid.extent[[1, 2, 3, 4]]
        else:
            print('not a direction')
        return _slice, extent

    def get_mask_surface_data(self, radius=None):
        points_interf = np.vstack(
            (self.model._surface_points.df['X'].values, self.model._surface_points.df['Y'].values)).T
        points_orient = np.vstack((self.model._orientations.df['X'].values, self.model._orientations.df['Y'].values)).T

        mask_interf = self.get_data_within_extent(points_interf)
        mask_orient = self.get_data_within_extent(points_orient)

        xj = self.model._grid.topography.values_3D[:, :, 0][0, :]
        yj = self.model._grid.topography.values_3D[:, :, 1][:, 0]
        zj = self.model._grid.topography.values_3D[:, :, 2].T

        interpolate = RegularGridInterpolator((xj, yj), zj)

        Z_interf_interp = interpolate(points_interf[mask_interf])
        Z_orient_interp = interpolate(points_orient[mask_orient])

        if radius is None:
            radius = np.diff(zj).max()
        print(radius)

        dist_interf = np.abs(Z_interf_interp - self.model._surface_points.df['Z'].values[mask_interf])
        dist_orient = np.abs(Z_orient_interp - self.model._orientations.df['Z'].values[mask_orient])

        surfmask_interf = dist_interf < radius
        surfmask_orient = dist_orient < radius
        surf_indexes = np.flatnonzero(mask_interf)[surfmask_interf]
        orient_indexes = np.flatnonzero(mask_orient)[surfmask_orient]

        mask_surfpoints = np.zeros(points_interf.shape[0], dtype=bool)
        mask_orient = np.zeros(points_orient.shape[0], dtype=bool)

        mask_surfpoints[surf_indexes] = True
        mask_orient[orient_indexes] = True

        return mask_surfpoints, mask_orient

    def get_mask_block_section(self, cell_number=3, direction='y', radius=None):
        if direction == 'x':
            column = 'X'
            start = self.model._grid.regular_grid.extent[0]
            end = self.model._grid.regular_grid.extent[1]
            r_o_inf = self.model._grid.regular_grid.dx
        elif direction == 'y':
            column = 'Y'
            start = self.model._grid.regular_grid.extent[2]
            end = self.model._grid.regular_grid.extent[3]
            r_o_inf = self.model._grid.regular_grid.dy
        elif direction == 'z':
            column = 'Z'
            start = self.model._grid.regular_grid.extent[4]
            end = self.model._grid.regular_grid.extent[5]
            r_o_inf = self.model._grid.regular_grid.dz
        else:
            raise

        if cell_number < 0:
            cell_number = end + cell_number + 1

        if radius is None:
            radius = r_o_inf
        coord = start + radius * cell_number
        mask_surfpoints = np.abs(self.model._surface_points.df[column].values - coord) < radius
        mask_orient = np.abs(self.model._orientations.df[column].values - coord) < radius
        return mask_surfpoints, mask_orient

    def get_mask_sections(self, j, radius=None):
        points_interf = np.vstack(
            (self.model._surface_points.df['X'].values, self.model._surface_points.df['Y'].values)).T
        points_orient = np.vstack((self.model._orientations.df['X'].values,
                                   self.model._orientations.df['Y'].values)).T
        if radius is None:
            radius = self.model._grid.sections.dist[j] / self.model._grid.sections.resolution[j][0]

        p1, p2 = np.array(self.model._grid.sections.points[j][0]), np.array(self.model._grid.sections.points[j][1])

        d_interf = np.abs(np.cross(p2 - p1, points_interf - p1) / np.linalg.norm(p2 - p1))
        d_orient = np.abs(np.cross(p2 - p1, points_orient - p1) / np.linalg.norm(p2 - p1))

        mask_surfpoints = d_interf < radius
        mask_orient = d_orient < radius
        return mask_surfpoints, mask_orient

    def get_data_within_extent(self, pts, ext=None):
        # ext = geo_model.grid.regular_grid.extent[:4]
        if ext is None:
            #extent must be the one of the topography (is half cell size smaller than regualar grid) because of
            #interpolation function
            ext = np.array([self.model._grid.topography.values_3D[:, :, 0][0, :][[0, -1]],
                            self.model._grid.topography.values_3D[:, :, 1][:, 0][[0, -1]]]).ravel()
        mask_x = np.logical_and(pts[:, 0] >= ext[0], pts[:, 0] <= ext[1])
        mask_y = np.logical_and(pts[:, 1] >= ext[2], pts[:, 1] <= ext[3])
        return np.logical_and(mask_x, mask_y)

    def plot_section_traces(self, section_names=None, contour_lines=True, show_data=True, show_all_data=False):
        if section_names is None:
            section_names = list(self.model._grid.sections.names)

        if show_data:
            if not show_all_data:
                self.plot_data(direction='z', at=section_names)
            else:
                self.plot_data(direction='z', at='everywhere')

        for section in section_names:
            j = np.where(self.model._grid.sections.names == section)[0][0]
            plt.plot([self.model._grid.sections.points[j][0][0], self.model._grid.sections.points[j][1][0]],
                     [self.model._grid.sections.points[j][0][1], self.model._grid.sections.points[j][1][1]],
                     label=section, linestyle='--')

            plt.xlim(self.model._grid.regular_grid.extent[:2])
            plt.ylim(self.model._grid.regular_grid.extent[2:4])
            # plt.set_aspect(np.diff(geo_model.grid.regular_grid.extent[:2])/np.diff(geo_model.grid.regular_grid.extent[2:4]))
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    def plot_stereonet(self, litho=None, planes=True, poles=True, single_plots=False,
                       show_density=False):
        try:
            import mplstereonet
        except ImportError:
            warnings.warn('mplstereonet package is not installed. No stereographic projection available.')

        from collections import OrderedDict
        import pandas as pn

        if litho is None:
            litho = self.model._orientations.df['surface'].unique()

        if single_plots is False:
            fig, ax = mplstereonet.subplots(figsize=(5, 5))
            df_sub2 = pn.DataFrame()
            for i in litho:
                df_sub2 = df_sub2.append(self.model._orientations.df[self.model._orientations.df['surface'] == i])

        for formation in litho:
            if single_plots:
                fig = plt.figure(figsize=(5, 5))
                ax = fig.add_subplot(111, projection='stereonet')
                ax.set_title(formation, y=1.1)

            # if series_only:
            # df_sub = self.model.orientations.df[self.model.orientations.df['series'] == formation]
            # else:
            df_sub = self.model._orientations.df[self.model._orientations.df['surface'] == formation]

            if poles:
                ax.pole(df_sub['azimuth'] - 90, df_sub['dip'], marker='o', markersize=10,
                        markerfacecolor=self._color_lot[formation],
                        markeredgewidth=1.1, markeredgecolor='gray', label=formation + ': ' + 'pole point')
            if planes:
                ax.plane(df_sub['azimuth'] - 90, df_sub['dip'], color=self._color_lot[formation],
                         linewidth=1.5, label=formation + ': ' + 'azimuth/dip')
            if show_density:
                if single_plots:
                    ax.density_contourf(df_sub['azimuth'] - 90, df_sub['dip'],
                                        measurement='poles', cmap='viridis', alpha=.5)
                else:
                    ax.density_contourf(df_sub2['azimuth'] - 90, df_sub2['dip'], measurement='poles', cmap='viridis',
                                        alpha=.5)

            fig.subplots_adjust(top=0.8)
            handles, labels = ax.get_legend_handles_labels()
            by_label = OrderedDict(zip(labels, handles))
            ax.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1.9, 1.1))
            ax._grid(True, color='black', alpha=0.25)


class PlotSolution(PlotData2D):

    def __init__(self, model):
        self.model = model
        # super().__init__(self)
        self._color_lot = dict(zip(self.model._surfaces.df['surface'], self.model._surfaces.df['color']))
        self._cmap = mcolors.ListedColormap(list(self.model._surfaces.df['color']))
        self._norm = mcolors.Normalize(vmin=0.5, vmax=len(self._cmap.colors) + 0.5)
        self._show_legend = False

    def plot_map(self, solution: Solution = None, contour_lines=False, show_data=True,
                 show_all_data=False, show_hillshades: bool = False, figsize=(12, 12), **kwargs):
        """

        Args:
            solution:
            contour_lines:
            show_data:
            show_all_data:
            show_hillshades: Calculate and add hillshading using elevation data
            figsize:
            **kwargs:
                - azdeg: float = azimuth of sun for hillshade
                - altdeg: float = altitude in degrees of sun for hillshade

        """
        azdeg: float = kwargs.get('azdeg', 315.0)
        altdeg: float = kwargs.get('altdeg', 45.0)
        if solution is None:
            solution = self.model.solutions

        if solution.geological_map is None:
            raise AttributeError('Geological map not computed. Activate the topography grid.')

        try:
            geomap = solution.geological_map[0].reshape(self.model._grid.topography.values_3D[:, :, 2].shape)
        except AttributeError:
            warnings.warn('Geological map not computed. Activate the topography grid.')

        if show_data == True and show_hillshades == True:
            self.plot_data(direction='z', at='topography', show_all_data=show_all_data)
            ls = LightSource(azdeg=azdeg, altdeg=altdeg)
            hillshade_topography = ls.hillshade(self.model.grid.topography.values_3D[:, :, 2])
            plt.imshow(hillshade_topography, origin='lower', extent=self.model.grid.topography.extent, alpha=0.75)
        elif show_data == True and show_hillshades == False:
            self.plot_data(direction='z', at='topography', show_all_data=show_all_data)
        else:
            fig, ax = plt.subplots(figsize=figsize)
        im = plt.imshow(geomap, origin='lower', extent=self.model._grid.topography.extent, cmap=self._cmap,
                        norm=self._norm, zorder=-100)

        if contour_lines == True and show_data == False:
            CS = ax.contour(self.model._grid.topography.values_3D[:, :, 2], cmap='Greys', linestyles='solid',
                            extent=self.model._grid.topography.extent, zlevel=200)
            ax.clabel(CS, inline=1, fontsize=10, fmt='%d')
            plothelp.add_colorbar(im=im, label='elevation [m]', cs=CS, aspect=35)

        # self.extract_section_lines('topography')

        plt.title("Geological map", fontsize=15)
        plt.xlabel('X')
        plt.ylabel('Y')

    def extract_section_lines(self, section_name=None, axes=None, zorder=2, faults_only=False):
        # Todo merge this with extract fault lines
        faults = list(self.model._faults.df[self.model._faults.df['isFault'] == True].index)
        if section_name == 'topography':
            shape = self.model._grid.topography.resolution
            a = self.model.solutions.geological_map[1]
            extent = self.model._grid.topography.extent
        else:
            l0, l1 = self.model._grid.sections.get_section_args(section_name)
            j = np.where(self.model._grid.sections.names == section_name)[0][0]
            shape = [self.model._grid.sections.resolution[j][0], self.model._grid.sections.resolution[j][1]]
            a = self.model.solutions.sections[1][:, l0:l1]
            extent = [0, self.model._grid.sections.dist[j],
                      self.model._grid.regular_grid.extent[4],
                      self.model._grid.regular_grid.extent[5]]

        counter = a.shape[0]
        if faults_only:
            counters = np.arange(0, len(faults), 1)
        else:
            counters = np.arange(0, counter, 1)

        c_id = 0  # color id startpoint
        for f_id in counters:
            block = a[f_id]
            level = self.model.solutions.scalar_field_at_surface_points[f_id][np.where(
                self.model.solutions.scalar_field_at_surface_points[f_id] != 0)]

            levels = np.insert(level, 0, block.max())
            c_id2 = c_id + len(level)
            if f_id == counters.max():
                levels = np.insert(levels, level.shape[0], block.min())
                c_id2 = c_id + len(levels)  # color id endpoint
            if section_name == 'topography':
                block = block.reshape(shape)
            else:
                block = block.reshape(shape).T

            zorder = zorder - (f_id + len(level))
            if axes is None:
                axes = plt.gca()

            if f_id >= len(faults):
                axes.contourf(block, 0, levels=np.sort(levels), colors=self._cmap.colors[c_id:c_id2][::-1],
                              linestyles='solid', origin='lower',
                              extent=extent, zorder=zorder)
            else:
                axes.contour(block, 0, levels=np.sort(levels), colors=self._cmap.colors[c_id:c_id2][0],
                             linestyles='solid', origin='lower',
                             extent=extent, zorder=zorder)
            c_id += len(level)

    def extract_fault_lines(self, cell_number=25, direction='y'):
        faults = list(self.model._faults.df[self.model._faults.df['isFault'] == True].index)
        if len(faults) == 0:
            pass
        else:
            _slice, extent = self._slice2D(cell_number, direction)
            for fault in faults:
                f_id = int(self.model._stack.df.loc[fault, 'order_series']) - 1
                block = self.model.solutions.scalar_field_matrix[f_id]
                level = self.model.solutions.scalar_field_at_surface_points[f_id][np.where(
                    self.model.solutions.scalar_field_at_surface_points[f_id] != 0)]
                level.sort()
                plt.contour(block.reshape(self.model._grid.regular_grid.resolution)[_slice].T, 0, extent=extent,
                            levels=level,
                            colors=self._cmap.colors[f_id], linestyles='solid')

    def plot_section_by_name(self, section_name, show_data=True, show_faults=True, show_topo=True,
                             show_all_data=False, contourplot=True, radius='default', **kwargs):

        if self.model.solutions.sections is None:
            raise AttributeError('no sections for plotting defined')
        if section_name not in self.model._grid.sections.names:
            raise AttributeError(f'Section "{section_name}" is not defined. '
                                 f'Available sections for plotting: {self.model._grid.sections.names}')

        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        l0, l1 = self.model._grid.sections.get_section_args(section_name)
        shape = self.model._grid.sections.resolution[j]

        image = self.model.solutions.sections[0][0][l0:l1].reshape(shape[0], shape[1]).T
        extent = [0, self.model._grid.sections.dist[j][0],
                  self.model._grid.regular_grid.extent[4], self.model._grid.regular_grid.extent[5]]

        if show_data:
            self.plot_section_data(section_name=section_name, show_all_data=show_all_data, radius=radius)

        axes = plt.gca()
        axes.imshow(image, origin='lower', zorder=-100,
                    cmap=self._cmap, norm=self._norm, extent=extent)
        if show_faults and not contourplot:
            self.extract_section_lines(section_name, axes, faults_only=True)
        else:
            self.extract_section_lines(section_name, axes, faults_only=False)
        if show_topo:
            if self.model._grid.topography is not None:
                alpha = kwargs.get('alpha', 1)
                xy = self.make_topography_overlay_4_sections(j)
                axes.fill(xy[:, 0], xy[:, 1], 'k', zorder=10, alpha=alpha)

        labels, axname = self._make_section_xylabels(section_name, len(axes.get_xticklabels()) - 1)
        pos_list = np.linspace(0, self.model._grid.sections.dist[j], len(labels))
        axes.xaxis.set_major_locator(FixedLocator(nbins=len(labels), locs=pos_list))
        axes.xaxis.set_major_formatter(FixedFormatter((labels)))
        axes.set(title=self.model._grid.sections.names[j], xlabel=axname, ylabel='Z')

    def plot_all_sections(self, show_data=False, section_names=None, show_topo=True,
                          figsize=(12, 12)):
        if self.model.solutions.sections is None:
            raise AttributeError('no sections for plotting defined')
        if self.model._grid.topography is None:
            show_topo = False
        if section_names is not None:
            if isinstance(section_names, list):
                section_names = np.array(section_names)
        else:
            section_names = self.model._grid.sections.names

        shapes = self.model._grid.sections.resolution
        fig, axes = plt.subplots(nrows=len(section_names), ncols=1, figsize=figsize)
        for i, section in enumerate(section_names):
            j = np.where(self.model._grid.sections.names == section)[0][0]
            l0, l1 = self.model._grid.sections.get_section_args(section)

            self.extract_section_lines(section, axes[i], faults_only=False)

            if show_topo:
                xy = self.make_topography_overlay_4_sections(j)
                axes[i].fill(xy[:, 0], xy[:, 1], 'k', zorder=10)

            # if show_data:
            #    section = str(section)
            #    print(section)
            #    self.plot_section_data(section_name=section)

            axes[i].imshow(self.model.solutions.sections[0][0][l0:l1].reshape(shapes[j][0], shapes[j][1]).T,
                           origin='lower', zorder=-100,
                           cmap=self._cmap, norm=self._norm, extent=[0, self.model._grid.sections.dist[j],
                                                                     self.model._grid.regular_grid.extent[4],
                                                                     self.model._grid.regular_grid.extent[5]])

            labels, axname = self._make_section_xylabels(section, len(axes[i].get_xticklabels()) - 1)
            pos_list = np.linspace(0, self.model._grid.sections.dist[j], len(labels))
            axes[i].xaxis.set_major_locator(FixedLocator(nbins=len(labels), locs=pos_list))
            axes[i].xaxis.set_major_formatter(FixedFormatter((labels)))
            axes[i].set(title=self.model._grid.sections.names[j], xlabel=axname, ylabel='Z')

        fig.tight_layout()

    def plot_section_scalarfield(self, section_name, sn, levels=50, show_faults=True, show_topo=True, lithback=True):
        if self.model.solutions.sections is None:
            raise AttributeError('no sections for plotting defined')
        if self.model._grid.topography is None:
            show_topo = False
        shapes = self.model._grid.sections.resolution
        fig = plt.figure(figsize=(16, 10))
        axes = fig.add_subplot(1, 1, 1)
        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        l0, l1 = self.model._grid.sections.get_section_args(section_name)
        if show_faults:
            self.extract_section_fault_lines(section_name, zorder=9)

        if show_topo:
            xy = self.make_topography_overlay_4_sections(j)
            axes.fill(xy[:, 0], xy[:, 1], 'k', zorder=10)

        axes.contour(self.model.solutions.sections[1][sn][l0:l1].reshape(shapes[j][0], shapes[j][1]).T,
                     # origin='lower',
                     levels=levels, cmap='autumn', extent=[0, self.model._grid.sections.dist[j],
                                                           self.model._grid.regular_grid.extent[4],
                                                           self.model._grid.regular_grid.extent[5]], zorder=8)
        axes.set_aspect('equal')
        if lithback:
            axes.imshow(self.model.solutions.sections[0][0][l0:l1].reshape(shapes[j][0], shapes[j][1]).T,
                        origin='lower',
                        cmap=self._cmap, norm=self._norm, extent=[0, self.model._grid.sections.dist[j],
                                                                  self.model._grid.regular_grid.extent[4],
                                                                  self.model._grid.regular_grid.extent[5]])

        labels, axname = self._make_section_xylabels(section_name, len(axes.get_xticklabels()))
        pos_list = np.linspace(0, self.model._grid.sections.dist[j], len(labels))
        axes.xaxis.set_major_locator(FixedLocator(nbins=len(labels), locs=pos_list))
        axes.xaxis.set_major_formatter(FixedFormatter((labels)))
        axes.set(title=self.model._grid.sections.names[j], xlabel=axname, ylabel='Z')

    def _slice_topo_4_sections(self, p1, p2, resx, resy):
        xy = self.model._grid.sections.calculate_line_coordinates_2points(np.array(p1),
                                                                          np.array(p2),
                                                                          resx)
        z = self.model._grid.topography.interpolate_zvals_at_xy(xy)
        return xy[:, 0], xy[:, 1], z

    def make_topography_overlay_4_sections(self, j):
        startend = list(self.model._grid.sections.section_dict.values())[j]
        p1, p2 = startend[0], startend[1]
        x, y, z = self._slice_topo_4_sections(p1, p2, self.model._grid.topography.resolution[0],
                                              self.model._grid.topography.resolution[1])
        pseudo_x = np.linspace(0, self.model._grid.sections.dist[j][0], z.shape[0])
        a = np.vstack((pseudo_x, z)).T
        a = np.append(a,
                      ([self.model._grid.sections.dist[j][0], a[:, 1][-1]],
                       [self.model._grid.sections.dist[j][0], self.model._grid.regular_grid.extent[5]],
                       [0, self.model._grid.regular_grid.extent[5]],
                       [0, a[:, 1][0]]))
        return a.reshape(-1, 2)

    def make_topography_overlay_4_blockplot(self, cell_number, direction):
        p1, p2 = self.calculate_p1p2(direction, cell_number)
        resx = self.model._grid.topography.resolution[0]
        resy = self.model._grid.topography.resolution[1]
        print('p1', p1, 'p2', p2)
        x, y, z = self._slice_topo_4_sections(p1, p2, resx, resy)
        if direction == 'x':
            a = np.vstack((y, z)).T
            ext = self.model._grid.regular_grid.extent[[2, 3]]
        elif direction == 'y':
            a = np.vstack((x, z)).T
            ext = self.model._grid.regular_grid.extent[[0, 1]]
        a = np.append(a,
                      ([ext[1], a[:, 1][-1]],
                       [ext[1], self.model._grid.regular_grid.extent[5]],
                       [ext[0], self.model._grid.regular_grid.extent[5]],
                       [ext[0], a[:, 1][0]]))
        line = a.reshape(-1, 2)
        plt.fill(line[:, 0], line[:, 1], color='k')

    def calculate_p1p2(self, direction, cell_number):
        if direction == 'y':
            y = self.model._grid.regular_grid.extent[2] + self.model._grid.regular_grid.dy * cell_number
            p1 = [self.model._grid.regular_grid.extent[0], y]
            p2 = [self.model._grid.regular_grid.extent[1], y]

        elif direction == 'x':
            x = self.model._grid.regular_grid.extent[0] + self.model._grid.regular_grid.dx * cell_number
            p1 = [x, self.model._grid.regular_grid.extent[2]]
            p2 = [x, self.model._grid.regular_grid.extent[3]]

        else:
            raise NotImplementedError
        return p1, p2

    def _make_section_xylabels(self, section_name, n=5):
        if n > 10:
            n = n - 2  # todo I don't know why but sometimes it wants to make a lot of xticks
        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        startend = list(self.model._grid.sections.section_dict.values())[j]
        p1, p2 = startend[0], startend[1]
        xy = self.model._grid.sections.calculate_line_coordinates_2points(p1,
                                                                          p2,
                                                                          n)
        if len(np.unique(xy[:, 0])) == 1:
            labels = xy[:, 1].astype(int)
            axname = 'Y'
        elif len(np.unique(xy[:, 1])) == 1:
            labels = xy[:, 0].astype(int)
            axname = 'X'
        else:
            labels = [str(xy[:, 0].astype(int)[i]) + ',\n' + str(xy[:, 1].astype(int)[i]) for i in
                      range(xy[:, 0].shape[0])]
            axname = 'X,Y'
        return labels, axname

    def plot_block_section(self, solution: Solution, cell_number: int, block: np.ndarray = None, direction: str = "y",
                           interpolation: str = 'none', show_data: bool = False, show_faults: bool = False,
                           show_topo: bool = False,
                           block_type=None, ve: float = 1, show_legend: bool = True, show_all_data: bool = False,
                           ax=None,
                           **kwargs):
        """Plot a section of the block model

        Args:
            solution (Solution): [description]
            cell_number (int): Section position of the array to plot.
            block (np.ndarray, optional): Lithology block. Defaults to None.
            direction (str, optional): Cartesian direction to be plotted
                ("x", "y", "z"). Defaults to "y".
            interpolation (str, optional): Type of interpolation of plt.imshow.
                Defaults to 'none'. Acceptable values are ('none' ,'nearest',
                'bilinear', 'bicubic', 'spline16', 'spline36', 'hanning',
                'hamming', 'hermite', 'kaiser', 'quadric', 'catrom', 'gaussian',
                'bessel', 'mitchell', 'sinc', 'lanczos'.
            show_data (bool, optional): Plots input data on-top of block
                section. Defaults to False.
            show_legend (bool, optional): Plot or hide legend - only available
                if no data is plotted.
            show_faults (bool, optional): Plot fault line on-top of block
                section. Defaults to False.
            show_topo (bool, optional): Plots block section with topography.
                Defaults to True.
            block_type ([type], optional): [description]. Defaults to None.
            ve (float, optional): Vertical exaggeration. Defaults to 1.

        Returns:
            (gempy.plot.visualization_2d.PlotData2D) Block section plot.
        """
        if block is None:
            _block = solution.lith_block
        else:
            _block = block
            if _block.dtype == bool:
                kwargs['cmap'] = 'viridis'
                kwargs['norm'] = None

        if block_type is not None:
            raise NotImplementedError

        plot_block = _block.reshape(self.model._grid.regular_grid.resolution)
        _a, _b, _c, extent_val, x, y = self._slice(direction, cell_number)[:-2]

        if show_data:
            if show_all_data:
                at = 'everywhere'
            else:
                at = 'block_section'
            self.plot_data(cell_number=cell_number, direction=direction, at=at, show_legend=show_legend)

        # TODO: plot_topo option - need fault_block for that

        # apply vertical exageration
        if direction in ("x", "y"):
            aspect = ve
        else:
            aspect = 1

        if 'cmap' not in kwargs:
            kwargs['cmap'] = self._cmap
        if 'norm' not in kwargs:
            kwargs['norm'] = self._norm

        sliced_block = plot_block[_a, _b, _c].T

        imshow_kwargs = kwargs.copy()
        if 'show_grid' in imshow_kwargs:
            imshow_kwargs.pop('show_grid')
        if 'grid_linewidth' in imshow_kwargs:
            imshow_kwargs.pop('grid_linewidth')

        im = plt.imshow(sliced_block,
                        origin="lower",
                        extent=extent_val,
                        interpolation=interpolation,
                        aspect=aspect,
                        **imshow_kwargs)

        if extent_val[3] < extent_val[2]:  # correct vertical orientation of plot
            plt.gca().invert_yaxis()  # if maximum vertical extent negative

        if show_faults:
            self.extract_fault_lines(cell_number, direction)

        if show_topo:
            if self.model._grid.topography is not None:
                if direction == 'z':
                    plt.contour(self.model._grid.topography.values_3D[:, :, 2], extent=extent_val, cmap='Greys')
                else:
                    self.make_topography_overlay_4_blockplot(cell_number=cell_number, direction=direction)

        if self._show_legend and show_legend:
            show_data = False  # to plot legend even when there are no data points in the section
        if not show_data and show_legend:
            patches = [mpatches.Patch(color=color, label=surface) for surface, color in self._color_lot.items()]
            plt.legend(handles=patches, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

        if 'show_grid' in kwargs:
            # TODO This only works fine for the y projection
            ax = plt.gca();
            ax.set_xticks(np.linspace(extent_val[0], extent_val[1], sliced_block.shape[0] + 1));
            ax.set_yticks(np.linspace(extent_val[2], extent_val[3], sliced_block.shape[1] + 1));

            grid_linewidth = kwargs.get('grid_linewidth', 1)
            ax._grid(color='w', linestyle='-', linewidth=grid_linewidth)

        plt.xlabel(x)
        plt.ylabel(y)
        return plt.gcf()

    def plot_scalar_field(self, solution, cell_number, series=0, N=20, block=None,
                          direction="y", alpha=0.6, show_data=True, show_all_data=False,
                          *args, **kwargs):
        """
        Plot a scalar field in a given direction.

        Args:
            cell_number(int): position of the array to plot
            scalar_field(str): name of the scalar field (or series) to plot
            n_pf(int): number of the  scalar field (or series) to plot
            direction(str): xyz. Caartesian direction to be plotted
            serie: *Deprecated*
            **kwargs: plt.contour kwargs

        Returns:
            scalar field plot
        """

        if isinstance(solution, Solution):
            scalar_field = solution.scalar_field_matrix[series]
        else:
            warnings.warn('Passing the block directly will get deprecated in the next version. Please use Solution'
                          'and block_type instead', FutureWarning)
            scalar_field = solution

        if 'cmap' not in kwargs:
            kwargs['cmap'] = 'magma'

        if show_data:
            if show_all_data:
                at = 'everywhere'
            else:
                at = 'block_section'
            self.plot_data(cell_number=cell_number, direction=direction, at=at)

        _a, _b, _c, extent_val, x, y = self._slice(direction, cell_number)[:-2]

        plt.contour(scalar_field.reshape(
            self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
            self.model._grid.regular_grid.resolution[2])[_a, _b, _c].T,
                    N,
                    extent=extent_val, *args,
                    **kwargs)

        plt.contourf(scalar_field.reshape(
            self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
            self.model._grid.regular_grid.resolution[2])[_a, _b, _c].T,
                     N,
                     extent=extent_val, alpha=alpha, *args,
                     **kwargs)

        if 'colorbar' in kwargs:
            plt.colorbar()

        plt.xlabel(x)
        plt.ylabel(y)

    @staticmethod
    def plot_topo_g(geo_model, edges, centroids, direction="y", scale=False,
                    label_kwargs=None, edge_kwargs=None):
        res = geo_model._grid.regular_grid.resolution
        if direction == "y":
            c1, c2 = (0, 2)
            e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
            e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
            d1 = geo_model._grid.regular_grid.extent[0]
            d2 = geo_model._grid.regular_grid.extent[4]
            # if len(list(centroids.items())[0][1]) == 2:
            #     c1, c2 = (0, 1)
            r1 = res[0]
            r2 = res[2]
        elif direction == "x":
            c1, c2 = (1, 2)
            e1 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
            e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
            d1 = geo_model._grid.regular_grid.extent[2]
            d2 = geo_model._grid.regular_grid.extent[4]
            # if len(list(centroids.items())[0][1]) == 2:
            #     c1, c2 = (0, 1)
            r1 = res[1]
            r2 = res[2]
        elif direction == "z":
            c1, c2 = (0, 1)
            e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
            e2 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
            d1 = geo_model._grid.regular_grid.extent[0]
            d2 = geo_model._grid.regular_grid.extent[2]
            # if len(list(centroids.items())[0][1]) == 2:
            #     c1, c2 = (0, 1)
            r1 = res[0]
            r2 = res[1]

        tkw = {
            "color": "white",
            "fontsize": 13,
            "ha": "center",
            "va": "center",
            "weight": "ultralight",
            "family": "monospace",
            "verticalalignment": "center",
            "horizontalalignment": "center",
            "bbox": dict(boxstyle='round', facecolor='black', alpha=1),
        }
        if label_kwargs is not None:
            tkw.update(label_kwargs)

        lkw = {
            "linewidth": 1,
            "color": "black"
        }
        if edge_kwargs is not None:
            lkw.update(edge_kwargs)

        for a, b in edges:
            # plot edges
            x = np.array([centroids[a][c1], centroids[b][c1]])
            y = np.array([centroids[a][c2], centroids[b][c2]])
            if scale:
                x = x * e1 / r1 + d1
                y = y * e2 / r2 + d2
            plt.plot(x, y, **lkw)

        for node in np.unique(list(edges)):
            x = centroids[node][c1]
            y = centroids[node][c2]
            if scale:
                x = x * e1 / r1 + d1
                y = y * e2 / r2 + d2
            plt.text(x, y, str(node), **tkw)

    def plot_gradient(self, scalar_field, gx, gy, gz, cell_number, quiver_stepsize=5,
                      # maybe call r sth. like "stepsize"?
                      direction="y", plot_scalar=True, *args, **kwargs):  # include plot data?
        """
            Plot the gradient of the scalar field in a given direction.

            Args:
                geo_data (gempy.DataManagement.InputData): Input data of the model
                scalar_field(numpy.array): scalar field to plot with the gradient
                gx(numpy.array): gradient in x-direction
                gy(numpy.array): gradient in y-direction
                gz(numpy.array): gradient in z-direction
                cell_number(int): position of the array to plot
                quiver_stepsize(int): step size between arrows to indicate gradient
                direction(str): xyz. Caartesian direction to be plotted
                plot_scalar(bool): boolean to plot scalar field
                **kwargs: plt.contour kwargs

            Returns:
                None
        """
        if direction == "y":
            if plot_scalar:
                self.plot_scalar_field(scalar_field, cell_number, direction=direction, plot_data=False)
            U = gx.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[::quiver_stepsize,
                cell_number, ::quiver_stepsize].T
            V = gz.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[::quiver_stepsize,
                cell_number, ::quiver_stepsize].T
            plt.quiver(self.model._grid.values[:, 0].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                            self.model._grid.regular_grid.resolution[2])[
                       ::quiver_stepsize, cell_number, ::quiver_stepsize].T,
                       self.model._grid.values[:, 2].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                            self.model._grid.regular_grid.resolution[2])[
                       ::quiver_stepsize, cell_number, ::quiver_stepsize].T, U, V, pivot="tail",
                       color='blue', alpha=.6)
        elif direction == "x":
            if plot_scalar:
                self.plot_scalar_field(scalar_field, cell_number, direction=direction, plot_data=False)
            U = gy.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[cell_number, ::quiver_stepsize,
                ::quiver_stepsize].T
            V = gz.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[cell_number, ::quiver_stepsize,
                ::quiver_stepsize].T
            plt.quiver(self.model._grid.values[:, 1].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                             self.model._grid.regular_grid.resolution[2])[cell_number,
                       ::quiver_stepsize, ::quiver_stepsize].T,
                       self.model._grid.values[:, 2].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                             self.model._grid.regular_grid.resolution[2])[cell_number,
                       ::quiver_stepsize, ::quiver_stepsize].T, U, V,
                       pivot="tail",
                       color='blue', alpha=.6)
        elif direction == "z":
            if plot_scalar:
                self.plot_scalar_field(scalar_field, cell_number, direction=direction, plot_data=False)
            U = gx.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[::quiver_stepsize, ::quiver_stepsize,
                cell_number].T
            V = gy.reshape(self.model._grid.regular_grid.resolution[0], self.model._grid.regular_grid.resolution[1],
                           self.model._grid.regular_grid.resolution[2])[::quiver_stepsize, ::quiver_stepsize,
                cell_number].T
            plt.quiver(self.model._grid.values[:, 0].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                             self.model._grid.regular_grid.resolution[2])[
                       ::quiver_stepsize, ::quiver_stepsize, cell_number].T,
                       self.model._grid.values[:, 1].reshape(self.model._grid.regular_grid.resolution[0],
                                                            self.model._grid.regular_grid.resolution[1],
                                                             self.model._grid.regular_grid.resolution[2])[
                       ::quiver_stepsize, ::quiver_stepsize, cell_number].T, U, V,
                       pivot="tail",
                       color='blue', alpha=.6)
        else:
            raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")



import os
import pytest
import gempy_legacy as gp
import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
from gempy_legacy.plot.vista import GemPyToVista

input_path = os.path.dirname(__file__) + '/../../notebooks/data'


@pytest.mark.skipif("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                    reason="Skipping this test on Travis CI.")
class TestVista:
    @pytest.fixture(scope='module')
    def vista_object_only_data(self, one_fault_model_no_interp):
        """
        Args:
            one_fault_model:
        """
        from gempy_legacy.plot.vista import GemPyToVista

        return GemPyToVista(one_fault_model_no_interp,
                            #plotter_type='background')
                            plotter_type='basic',  off_screen=True)

    @pytest.fixture(scope='module')
    def vista_object_only_data_interactive(self, one_fault_model_no_interp):
        """
        Args:
            one_fault_model:
        """
        from gempy_legacy.plot.vista import GemPyToVista

        return GemPyToVista(one_fault_model_no_interp,
                            plotter_type='background')

    @pytest.fixture(scope='module')
    def vista_object_computed(self, one_fault_model_solution):
        """
        Args:
            one_fault_model_solution:
        """
        from gempy_legacy.plot.vista import GemPyToVista

        return GemPyToVista(one_fault_model_solution, plotter_type='basic',
                            off_screen=True
                            )

    @pytest.fixture(scope='module')
    def vista_object_computed_topo(self, one_fault_model_solution):
        """
        Args:
            one_fault_model_solution:
        """

        from gempy_legacy.plot.vista import GemPyToVista
        one_fault_model_solution.update_additional_data()
        one_fault_model_solution.update_to_interpolator()
        one_fault_model_solution.set_topography()
        gp.compute_model(one_fault_model_solution)

        return GemPyToVista(one_fault_model_solution, plotter_type='basic', off_screen=True)

    def test_set_bounds(self, vista_object_only_data):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data.set_bounds()

    def test_select_surface_points(self, vista_object_only_data):
        """
        Args:
            vista_object_only_data:
        """
        sp = vista_object_only_data._select_surfaces_data(data_df=vista_object_only_data.model._surface_points.df,
                                                          surfaces='all')
        np.testing.assert_almost_equal(sp.loc[4, 'X_c'],  0.486942, 5)

        sp2 = vista_object_only_data._select_surfaces_data(data_df=vista_object_only_data.model._surface_points.df,
                                                           surfaces=['Sandstone_2'])
        with pytest.raises(KeyError):
            sp2.loc[4, 'X_r']

    @pytest.mark.skip(reason= "github actions does not like background plot")
    def test_plot_surface_points_poly_live(self, vista_object_only_data_interactive):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data_interactive.live_updating = True
        vista_object_only_data_interactive.plot_surface_points()

    def test_plot_surface_points_poly_static(self, vista_object_only_data):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data.live_updating = False
        vista_object_only_data.plot_surface_points()
        img = vista_object_only_data.p.show(screenshot=True, auto_close=False)
        img = vista_object_only_data.p.last_image
        plt.imshow(img[1])
        plt.show(block=False)
        print('foo')

    def test_plot_surface_points_poly_clear(self, vista_object_only_data):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data.live_updating = False
        vista_object_only_data.plot_surface_points()
        vista_object_only_data.plot_surface_points()
        img = vista_object_only_data.p.show(screenshot=True, auto_close=False)
        img = vista_object_only_data.p.last_image
        plt.imshow(img[1])
        plt.show(block=False)
        print('foo')

    def test_plot_data_static(self, vista_object_only_data):
        vista_object_only_data.plot_data()
        img = vista_object_only_data.p.show(screenshot=True, auto_close=False)
        img = vista_object_only_data.p.last_image
        plt.imshow(img[1])
        plt.show()
        print('foo')

    @pytest.mark.skip(reason="github actions does not like background plot")
    def test_plot_orientations_poly_live(self, vista_object_only_data_interactive):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data_interactive.live_updating = True
        vista_object_only_data_interactive.plot_orientations()

        print('foo')

    def test_plot_orientations_poly_static(self, vista_object_only_data):
        """
        Args:
            vista_object_only_data:
        """
        vista_object_only_data.live_updating = False
        vista_object_only_data.plot_orientations()
        img = vista_object_only_data.p.show(screenshot=True, auto_close=False)
        img = vista_object_only_data.p.last_image
        plt.imshow(img[1])
        plt.show()

        print('foo')

    def test_plot_surfaces(self, vista_object_computed):
        """
        Args:
            vista_object_computed:
        """
        a = vista_object_computed.plot_surfaces()
        img = vista_object_computed.p.show(screenshot=True, auto_close=False)
        img = vista_object_computed.p.last_image
        plt.imshow(img[1])
        plt.show()

    def test_plot_surfaces_data(self, vista_object_computed):
        """
        Args:
            vista_object_computed:
        """
        a = vista_object_computed.plot_surfaces()
        vista_object_computed.plot_data()
        img = vista_object_computed.p.show(screenshot=True, auto_close=False)
        img = vista_object_computed.p.last_image
        plt.imshow(img[1])
        plt.show()

    def test_plot_topography_lith(self, one_fault_model_topo_solution):
        """
        Args:
            vista_object_computed_topo:
        """

        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_topography(scalars='geomap')
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()

    def test_plot_topography_high(self, one_fault_model_topo_solution):
        """
        Args:
            vista_object_computed_topo:
        """
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_topography(scalars='topography')
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()
        print('foo')

    def test_plot_topography_high_clear(self, one_fault_model_topo_solution):
        """
        Args:
            vista_object_computed_topo:
        """
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_topography(scalars='topography')
        one_fault_model_topo_solution.set_topography()
        gp.compute_model(one_fault_model_topo_solution)

        vista_object_computed_topo.plot_topography(scalars='topography')
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()
        print('foo')


    def test_plot_regular_grid_lith(self, vista_object_computed):
        """
        Args:
            vista_object_computed:
        """
        vista_object_computed.plot_structured_grid('lith', render_topography=False,
                                                   opacity=.8)
        img = vista_object_computed.p.show(screenshot=True)
        img = vista_object_computed.p.last_image
        plt.imshow(img[1])
        plt.show()

        print('foo')

    def test_plot_regular_grid_scalar_0(self, vista_object_computed):
        # Add all scalar fields to the pyvista object and plot lith
        """
        Args:
            vista_object_computed:
        """
        vista_object_computed.plot_structured_grid('all', render_topography=False)

        # Change active scalar to stratigraphy
        vista_object_computed.set_active_scalar_fields(
            'sf_Strat_Series')

        # Change the color map to the lithology cmap
        vista_object_computed.set_scalar_field_cmap('lith', vista_object_computed.regular_grid_actor)

        # Set active scalar back to lith
        vista_object_computed.set_active_scalar_fields('lith')

        print('foo')

    def test_plot_regular_grid_scalar_topo(self, one_fault_model_topo_solution):
        """
        Args:
            vista_object_computed_topo:
        """
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_structured_grid('lith', render_topography=False)
        vista_object_computed_topo.plot_surfaces()
        # vista_object_computed_topo.set_scalar_bar()
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()
        print('foo')

    def test_plot_regular_grid_scalar_1(self, vista_object_computed):
        """
        Args:
            vista_object_computed:
        """
        vista_object_computed.plot_structured_grid('scalar', series='Strat_Series', render_topography=True)
        print('foo')

    def test_plot_regular_grid_select_field(self, one_fault_model_solution):
        """
        Args:
            vista_object_computed:
        """

        vista_object_computed = GemPyToVista(one_fault_model_solution,
                                             plotter_type='basic',
                                             off_screen=True
                                           )
        vista_object_computed.plot_structured_grid('lith')
        with pytest.raises(KeyError):
            vista_object_computed.set_active_scalar_fields(scalar_field='scalar')
        # vista_object_computed.plot_structured_grid('scalar')
        print('foo')

    def test_plot_several(self, one_fault_model_topo_solution):
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_surface_points()
        vista_object_computed_topo.plot_surfaces()
        vista_object_computed_topo.plot_structured_grid('scalar',
                                                        series='Strat_Series',
                                                        render_topography=False)
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()

    def test_plot_several2(self, one_fault_model_topo_solution):
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_surface_points()
        vista_object_computed_topo.plot_surfaces()
        vista_object_computed_topo.plot_structured_grid(render_topography=False)
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()

    def test_plot_several3(self, one_fault_model_topo_solution):
        vista_object_computed_topo = GemPyToVista(one_fault_model_topo_solution,
                                                  plotter_type='basic', off_screen=True)

        vista_object_computed_topo.plot_surface_points()
        vista_object_computed_topo.plot_surfaces()
        vista_object_computed_topo.plot_structured_grid(render_topography=True,
                                                        scalar_field='scalar',
                                                        series='Strat_Series',
                                                        )
        vista_object_computed_topo.plot_topography(scalars='topography')
        img = vista_object_computed_topo.p.show(screenshot=True)
        img = vista_object_computed_topo.p.last_image
        plt.imshow(img[1])
        plt.show()



import os
import requests
import json
import datetime
try:
    import pyqrcode as qr
    PYQRCODE_IMPORT = True
except ImportError:
    PYQRCODE_IMPORT = False

package_directory = os.path.dirname(os.path.abspath(__file__))


class RexAPI:

    def __init__(self, project_name, api_token=None, secret=None):
        """Rest client to connect to the RexOS system


        Args:
            project_name:
            api_token:
            secret:
        """
        self.response = None  # saves the most current server response

        self.token_ID, self.secret = self.load_credentials(api_token=api_token, secret=secret)
        self.access_token = self.authorize_session()
        self.owner = self.get_user_information()
        self.project_name = project_name
        self.project_urn, self.project_link = self.create_project(self.project_name)
        self.root_reference_link, self.root_reference_key = self.create_root_reference()
        self.file_reference_link = self.create_file_resource_reference()

        self.project_file_urn = None
        self.project_file_uplink = None
        self.project_reference = None

        self.rextag = None

    def load_credentials(self, filename=os.path.join(package_directory, 'RexCloud_Api_key.txt'),
                         api_token: str = None, secret: str = None):

        if not os.path.isfile(filename) or (api_token is not None and secret is not None):
            file = open(filename, 'w')
            login_data = False
            if api_token is None or secret is None:
                login_data = True

            if api_token is None:
                api_token = 'REPLACE_TEXT_WITH_YOUR_API_Token'
            if secret is None:
                secret = 'REPLACE_TEXT_WITH_SECRET'

            file.write(api_token+'\n'+secret+'\n'+' # put your API tokens and secrets in the lines above.\n'
                                                  ' # Do not track the file on git.')
            file.close()

            if login_data:
                raise AttributeError('Cache key is not created. You need to pass as argument the REX api_token'
                                     ' and secret, or adding them in RexCloud_API_key.txt.'
                                     ' https://www.rexos.org/getting-started/')

        with open(filename, "r") as credential_file:
            token_id = credential_file.readline().strip('\n')
            secret = credential_file.readline().strip('\n')

        return token_id, secret

    def authorize_session(self):

        headers = {'Accept': 'application/json;charset=UTF-8',
                   'Content-Type': 'application/x-www-form-urlencoded; charset=ISO-8859-1'}

        data = {'grant_type': 'client_credentials'}

        self.response = requests.post('https://rex.robotic-eyes.com/oauth/token',
                                      headers=headers, data=data, auth=(self.token_ID, self.secret)
                                      )
        if self.response.status_code == 200:
            access_token = self.response.json()['access_token']
            return access_token

        else:
            raise ConnectionError("something went wrong! Status code: "+str(self.response.status_code) +
                                  'Probably the token or the secret is not valid. '
                                  'https://www.rexos.org/getting-started/')

    def get_user_information(self):
        headers = {'Authorization': 'Bearer ' + self.access_token, 'Accept': 'application/json;charset=UTF-8'}
        self.response = requests.get('https://rex.robotic-eyes.com/api/v2/users/current', headers=headers)

        if self.response.status_code == 200:
            owner = self.response.json()['userId']
            return owner

        else:
            print("something went wrong! Status code: "+str(self.response.status_code))

    def create_project(self, project_name):
        headers = {
            'Authorization': 'Bearer ' + self.access_token,
            'Accept': 'application/json;charset=UTF-8',
            'Content-Type': 'application/json;charset=UTF-8'
        }

        data = json.dumps({"name" : project_name,  "owner" : self.owner})  # this call expects json!

        self.response = requests.post('https://rex.robotic-eyes.com/api/v2/projects', headers=headers, data=data)

        if self.response.status_code == 201:
            project_urn = self.response.json()['urn']
            project_link = self.response.json()['_links']['self']['href']

            return project_urn, project_link

        else:
            raise ConnectionError("something went wrong! Status code: " + str(self.response.status_code) +
                                  'Probably project name already exists.')

    def create_root_reference(self):
        headers = {
            'Authorization': 'Bearer ' + self.access_token,
            'Accept': 'application/json;charset=UTF-8',
            'Content-Type': 'application/json;charset=UTF-8',
        }

        data = json.dumps({"project" : self.project_link,
                "name" : "root reference",
                "address" : {"addressLine1": "Sample", "postcode": "52072",
                             "city": "Aachen",   "country" : "Austria"}}
                          )

        self.response = requests.post('https://rex.robotic-eyes.com/api/v2/rexReferences', headers=headers, data=data)
        if self.response.status_code == 201:
            root_reference_link = self.response.json()['_links']['self']['href']
            root_reference_key = self.response.json()['key']

            return root_reference_link, root_reference_key

        else:
            print("something went wrong! Status code: " + str(self.response.status_code))

    def create_file_resource_reference(self):
        headers = {
            'Authorization': 'Bearer ' + self.access_token,
            'Accept': 'application/json;charset=UTF-8',
            'Content-Type': 'application/json;charset=UTF-8',
        }

        data = json.dumps({"project" : self.project_link,
                           "name" : "file ressource reference",
                           "rootReference" : "false",  #setting root reference to false
                           "parentReference" : self.root_reference_link})

        self.response = requests.post('https://rex.robotic-eyes.com/api/v2/rexReferences', headers=headers, data=data)
        if self.response.status_code == 201:
            file_reference_link = self.response.json()['_links']['self']['href']

            return file_reference_link

        else:
            print("something went wrong! Status code: " + str(self.response.status_code))

    def create_project_file(self, projectname):
        headers = {
            'Authorization':'Bearer ' + self.access_token,
            'Accept': 'application/json;charset=UTF-8',
            'Content-Type': 'application/json;charset=UTF-8',
                 }

        data = json.dumps({"project" : self.project_link,
                           "name" : projectname,
                           "type" : "rex",

                           "rexReference" : self.file_reference_link
                           })

        self.response = requests.post('https://rex.robotic-eyes.com/api/v2/projectFiles', headers=headers, data=data)

        if self.response.status_code == 201:
            self.project_file_urn = self.response.json()['urn']
            self.project_file_uplink = self.response.json()['_links']['file.upload']['href']
            self.project_reference = self.response.json()['rexReferenceKey']

            return True

        else:
            print("something went wrong! Status code: " + str(self.response.status_code))

            return False

    def upload_rexfile(self, filename):

        headers = {
            'Authorization': 'Bearer ' + self.access_token,
            'contentType': 'application/octet-stream',
            'type' : 'rex'
        }

        file = {
            'file':  (filename, open(filename, 'rb')),

        }

        self.response = requests.post(self.project_file_uplink,
                                 headers=headers, files=file)

    def return_rextag(self):
        self.rextag = Rextag(self.project_reference)

        return self.rextag


class Rextag:

    def __init__(self, project_reference):
        self.rextag_url, self.rextag = self.create_rextag(project_reference)

    def __repr__(self):
        return self.rextag.terminal(module_color="reverse", background="default", quiet_zone=1)

    def create_rextag(self, project_reference):
        if PYQRCODE_IMPORT is False:
            raise ImportError('This method depends on pyqrcode and it is not possible to import.')
        base_url = "https://rex.codes/v1/"
        rextag_url = base_url+project_reference
        rextag = qr.create(rextag_url)
        return rextag_url, rextag

    def display_tag(self, reverse=True):
        """
        displays the rextag to the terminal standard output as ascii.

        you can invert the color by setting inverse=True, this is necessary if you run it in a jupyter notebook
        after creation, you can save the rextag as svg using the rextag.svg method:
        self.rextag.svg("project_name", scale=8)

        Args:
            reverse: (boolean) inverts background and foreground color in the terminal output

        Returns: None, prints a QR code to terminal output

        """

        if reverse:
            print(self.rextag.terminal(module_color="reverse", background="default", quiet_zone=1))

        else:
            print(self.rextag.terminal(quiet_zone=1))

        return True

    def save_svg(self, filename):
        self.rextag.svg(filename, scale=8)


def upload_to_rexcloud(infiles : list, project_name=None, **kwargs):
    """
    wrapper around api calls to upload rexfiles of a gempy model.

    you will need to register an account under https://app.rexos.cloud/ .
    create an api key and store the key and secret in RexCloud_Api_key.txt.

    the function will take a list of rexos input filenames and uploads them into a newly created project.

    an ar code is plotted that can be scanned with the rexview app to show the model in vr.
    the qr code is plotted in ascii to the standard output, if the qr code is not recognizable,
     try to revert the colors o0f the terminal output by setting reverse=False in thew show_tag() Method.

    all api calls are python implementation of the Rex os api:
    https://www.rexos.org/rex-api/#tutorial-rex-project-before-you-begin


    Args:
        infiles: List of rexos file names
        project_name: name of the project under which it appears in the rexcloud. if none is specified,
        the current timestamp is used.

    Returns:
        A Rextag Object

    """

    if project_name is None:
        timestamp = datetime.datetime.now()
        project_name = str(timestamp)

    api = RexAPI(project_name, **kwargs)

    for file in infiles:

        api.create_project_file(file)
        api.upload_rexfile(file)

    tag = api.return_rextag()

    return tag





"""
Moureze
~~~~~~~

"""

# %% 
# These two lines are necessary only if gempy is not installed
import sys, os
os.environ["aesara_FLAGS"] = "mode=FAST_RUN,device=cpu"

# Importing gempy
import gempy_legacy as gp

# Aux imports
import numpy as np
import pandas as pn
import matplotlib.pyplot as plt


# %%
# Loading surface points from repository:
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 
# With pandas we can do it directly from the web and with the right args
# we can directly tidy the data in gempy style:
# 

# %% 
Moureze_points = pn.read_csv(
    'https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Moureze/Moureze_Points.csv', sep=';',
    names=['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', '_'], header=0, )
Sections_EW = pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Moureze/Sections_EW.csv',
                          sep=';',
                          names=['X', 'Y', 'Z', 'ID', '_'], header=1).dropna()
Sections_NS = pn.read_csv('https://raw.githubusercontent.com/Loop3D/ImplicitBenchmark/master/Moureze/Sections_NS.csv',
                          sep=';',
                          names=['X', 'Y', 'Z', 'ID', '_'], header=1).dropna()

# %%
# Extracting the orientatins:
# 

# %% 
mask_surfpoints = Moureze_points['G_x'] < -9999
surfpoints = Moureze_points[mask_surfpoints]
orientations = Moureze_points[~mask_surfpoints]

# %%
# Giving an arbitrary value name to the surface
# 

# %% 
surfpoints['surface'] = '0'
orientations['surface'] = '0'

# %% 
surfpoints.tail()

# %% 
orientations.tail()

# %%
# Data initialization:
# ~~~~~~~~~~~~~~~~~~~~
# 
# Suggested size of the axis-aligned modeling box:
# 
# Origin: -5 -5 -200
# 
# Maximum: 305 405 -50
# 
# Suggested resolution: 2m (grid size 156 x 206 x 76)
# 


# %%
# Only using one orientation because otherwhise it gets a mess
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 

# %% 
# Number voxels
np.array([156, 206, 76]).prod()

# %% 
resolution_requ = [156, 206, 76]
resolution = [77, 103, 38]
resolution_low = [45, 51, 38]
geo_model = gp.create_model('Moureze')
geo_model = gp.init_data(geo_model,
                         extent=[-5, 305, -5, 405, -200, -50], resolution=resolution_low,
                         surface_points_df=surfpoints, orientations_df=orientations,
                         surface_name='surface',
                         add_basement=True)

# %%
# Now we can see how the data looks so far:
# 

# %% 
gp.plot_2d(geo_model, direction='y')

# %% 
gp.set_interpolator(geo_model,
                    aesara_optimizer='fast_run', dtype='float64')

# %%
# The default range is always the diagonal of the extent. Since in this
# model data is very close we will need to reduce the range to 5-10% of
# that value:
# 

# %% 

new_range = geo_model.get_additional_data().loc[('Kriging', 'range'), 'values'] * 0.2
geo_model.modify_kriging_parameters('range', new_range)

# %%
gp.compute_model(geo_model, set_solutions=True, sort_surfaces=False)

# %%
# Time
# ~~~~
# 
# 300k voxels 3.5k points
# ^^^^^^^^^^^^^^^^^^^^^^^
# 
# -  Nvidia 2080: 500 ms ± 1.3 ms per loop (mean ± std. dev. of 7 runs, 1
#    loop each), Memory 1 Gb
# -  CPU 14.2 s ± 82.4 ms per loop (mean ± std. dev. of 7 runs, 1 loop
#    each), Memory: 1.3 Gb
# 
# 2.4 M voxels, 3.5k points
# ^^^^^^^^^^^^^^^^^^^^^^^^^
# 
# -  CPU 2min 33s ± 216 ms per loop (mean ± std. dev. of 7 runs, 1 loop
#    each) Memory: 1.3 GB
# -  Nvidia 2080: 1.92 s ± 6.74 ms per loop (mean ± std. dev. of 7 runs, 1
#    loop each) 1 Gb
# 
# 2.4 M voxels, 3.5k points 3.5 k orientations
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
# -  Nvidia 2080: 2.53 s ± 1.31 ms per loop (mean ± std. dev. of 7 runs, 1
#    loop each)
# 

# %% 

gp.plot_2d(geo_model, cell_number=[16], series_n=0, show_scalar=True)
# %% 
gp.plot_2d(geo_model, cell_number=16, show_data=True, direction='y')

# %%
# sphinx_gallery_thumbnail_number = 4
gp.plot_3d(geo_model)


# %%
# |image0|
# 
# .. |image0| image:: ./Moureze.png
# 
# 


# %%
# Export data:
# ~~~~~~~~~~~~
# 
# The solution is stored in a numpy array of the following shape. Axis 0
# are the scalar fields of each correspondent series/faults in the
# following order (except basement):
# 

# %% 
geo_model.series

# %%
# For the surfaces, there are two numpy arrays, one with vertices and the
# other with triangles. Axis 0 is each surface in the order:
# 

# %% 
geo_model.surfaces


# %%
# np.save('Moureze_scalar', geo_model.solutions.scalar_field_matrix)
# np.save('Moureze_ver', geo_model.solutions.vertices)
# np.save('Moureze_edges', geo_model.solutions.edges)
# gp.plot.export_to_vtk(geo_model, 'Moureze')

gp.save_model(geo_model)


"""
This file is part of gempy.

gempy is free software. For license details, see the LICENSE.md file supplied
 with gempy

Created on 10/10 /2018

@author: Miguel de la Varga
"""

import numpy as np
import pandas as pn
from numpy import ndarray
from typing import Union
import warnings
from gempy_legacy.core.model import Project
from gempy_legacy.core.solution import Solution
from gempy_legacy.utils.meta import _setdoc, _setdoc_pro


# region get
def get():
    pass


# endregion

# region edit
def edit(model: Project,  method: str, data_object: str = None, **kwargs):
    """Function to edit any of the data_objects of gempy. Check
        https://www.gempy.org/documentation data for documentation.

    Args:
        model (Project):
        data_object (str): Object that you want to edit
        method (str): Method you want to use
        **kwargs:
    """
    if data_object is not None:
        data_object = getattr(model, '_' + data_object)
    else:
        data_object = model
    m = getattr(data_object, method)
    return m(**kwargs)


# endregion


# region mapping
@_setdoc_pro(Project.__doc__)
def map_series_to_surfaces(geo_model: Project, mapping_object: Union[dict, pn.Categorical] = None,
                           set_series=True, sort_geometric_data: bool = True, remove_unused_series=True):
    """Mapping which surfaces belongs to which geological feature

    Args:
        geo_model (:class:`gempy.core.model.Project`): [s0]
        mapping_object: [s_mapping_object]
        set_series (bool): If True set missing series from the mapping object
        sort_geometric_data (bool): If True sort the geometric_data according to the new
         order of the stack
        remove_unused_series(bool):

    Returns
        :class:`gempy.core.data.Surfaces`
    """
    warnings.warn('Series is going to get renamed to Stack. Please use'
                  '`map_stack_to_surfaces` instead.', DeprecationWarning)

    geo_model.map_stack_to_surfaces(mapping_object, set_series, sort_geometric_data, remove_unused_series)
    return geo_model._surfaces


@_setdoc_pro(Project.__doc__)
def map_stack_to_surfaces(geo_model: Project, mapping_object: Union[dict, pn.Categorical] = None,
                          set_features=True, sort_geometric_data: bool = True, remove_unused_series=True):
    """Mapping which surfaces belongs to which geological feature

    Args:
        geo_model (:class:`gempy.core.model.Project`): [s0]
        mapping_object: [s_mapping_object]
        set_features (bool): If True set missing features from the mapping object
        sort_geometric_data (bool): If True sort the geometric_data according to the new
         order of the stack
        remove_unused_series(bool):

    Returns
        :class:`gempy.core.data.Surfaces`
    """

    geo_model.map_stack_to_surfaces(mapping_object, set_features, sort_geometric_data, remove_unused_series)
    return geo_model._surfaces

# endregion


# region create
def create_model(project_name='default_project') -> Project:
    """Create a Project object.

    Args:
        project_name (str): Name of the project

    Returns:
        :class:`gempy.core.model.Project`

    See Also:
        :class:`gempy.core.model.Project`

    Notes:
        TODO: Adding saving address
    """
    return Project(project_name)


def create_data(project_name: str = 'default_project',
                extent: Union[list, ndarray] = None,
                resolution: Union[list, ndarray] = None,
                **kwargs) -> Project:
    """Create a :class:`gempy.core.model.Project` object and initialize some of
    the main functions such as:

    - Grid :class:`gempy.core.data.GridClass`: To regular grid.
    - read_csv: SurfacePoints and orientations: From csv files
    - set_values to default



    Args:
        project_name (str):
        extent (list or array): [x_min, x_max, y_min, y_max, z_min, z_max].
            Extent for the visualization of data and default of for the grid
            class.
        resolution (list or array): [nx, ny, nz]. Resolution for the
            visualization of data and default of for the grid class.
        **kwargs:

    Keyword:
        path_i: Path to the data bases of surface_points. Default os.getcwd(),
        path_o: Path to the data bases of orientations. Default os.getcwd()

    Returns:
        :class:`Project`
    """

    geo_model = create_model(project_name)
    return init_data(geo_model, extent=extent, resolution=resolution,
                     project_name=project_name, **kwargs)


# TODO We need to decide how to initialize a model. Having create_data and init
#  data  does not seem too robust
@_setdoc_pro([Project.__doc__])
def init_data(geo_model: Project, extent: Union[list, ndarray] = None,
              resolution: Union[list, ndarray] = None,
              **kwargs) -> Project:
    """Initialize some of the main functions such as:

     - Regular grid (:class:`gempy.core.data.Grid`).
     - read_csv: :class:`gempy.core.data_modules.geometric_data.SurfacePoints`
       and :class:`gempy.core.data_modules.geometric_data.Orientations` From csv files
     - set_values to default

    Args:
        geo_model (Project): [s0]
        extent: [s_extent]
        resolution: [s_resolution]

    Keyword Args:
        path_i: [s_path_i]
        path_o: [s_path_o]
        surface_points_df: [s_surface_points_df]
        orientations_df: [s_orientations_df]
    Returns:
        :class:`gempy.core.model.Project`
    """

    if extent is None or resolution is None:
        warnings.warn('Regular grid won\'t be initialized, you will have to create a grid afterwards. See gempy.set_grid')
    else:
        geo_model.set_regular_grid(extent, resolution)

    if 'path_i' in kwargs or 'path_o' in kwargs:
        read_csv(geo_model, **kwargs)

    if 'surface_points_df' in kwargs:
        geo_model.set_surface_points(kwargs['surface_points_df'], **kwargs)
        # if we set the surfaces names with surfaces they cannot be set again on orientations or pandas will complain.
        kwargs['update_surfaces'] = False
    if 'orientations_df' in kwargs:
        geo_model.set_orientations(kwargs['orientations_df'], **kwargs)

    return geo_model


# endregion


# If everything works out properly update. Should happen automatically. Therefore
# we will keep it just as a method
# region update
def update_additional_data(model: Project, update_structure=True, update_kriging=True):
    """
    Args:
        model (Project):
        update_structure:
        update_kriging:
    """
    warnings.warn('This function is going to be deprecated. Use Project.update_additional_data instead',
                  DeprecationWarning)
    return model.update_additional_data(update_structure, update_kriging)


# endregion


# region io
# @_setdoc([Project.read_data.__doc__])
def read_csv(geo_model: Project, path_i=None, path_o=None, **kwargs):
    """
    Args:
        geo_model (Project):
        path_i:
        path_o:
        **kwargs:
    """
    if path_i is not None or path_o is not None:
        try:
            geo_model.read_data(path_i, path_o, **kwargs)
        except KeyError as e:
            raise KeyError('Loading of CSV file failed. Check if you use commas '
                           'to separate your data.' + str(e))
    return True


# endregion


# region Computing the model
@_setdoc_pro([Project.__doc__, Solution.compute_marching_cubes_regular_grid.__doc__,
              Project.set_surface_order_from_solution.__doc__])
def compute_model(model: Project, output=None, at: np.ndarray = None, compute_mesh=True,
                  reset_weights=False, reset_scalar=False,
                  reset_block=False, sort_surfaces=True,
                  debug=False, set_solutions=True,
                  **kwargs) -> Solution:
    """Computes the geological model and any extra output given in the
    additional data option.

    Args:
        model (Project): [s0]
        output (str {'geology', 'gravity'}): Compute the lithologies or gravity
        at (np.ndarray):
        compute_mesh (bool): if True compute marching cubes: [s1]
        reset_weights (bool): Not Implemented
        reset_scalar (bool): Not Implemented
        reset_block (bool): Not Implemented
        sort_surfaces (bool): if True call
            Project.set_surface_order_from_solution: [s2]
        debug (bool): if True, the computed interpolation are not stored in any
            object but instead returned
        set_solutions (bool): Default True. If True set the results into the
            :class:`Solutions` linked object.
        **kwargs:

    Keyword Args:
        compute_mesh_options (dict): options for the marching cube function. 1)
            rescale: True

    Returns:
        :class:`Solutions`
    """

    # Check config
    # ------------
    _check_valid_model_input(model)

    if output is not None:
        warnings.warn('Argument output has no effect anymore and will be deprecated in GemPy 2.2.'
                      'Set the output only in gempy.set_interpolator.', DeprecationWarning, )
    if at is not None:
        model._grid.deactivate_all_grids()
        model.set_custom_grid(at)

    # ------------

    i = model._interpolator.get_python_input_block(append_control=True, fault_drift=None)
    model._interpolator.reset_flow_control_initial_results(reset_weights, reset_scalar, reset_block)

    sol = model._interpolator.aesara_function(*i)

    if debug is True or set_solutions is False:
        return sol

    elif set_solutions is True:

        model.solutions.set_solutions(
            sol,
            compute_mesh,
            sort_surfaces,
            **kwargs)

        if sort_surfaces:
            model.set_surface_order_from_solution()
        return model.solutions


def _check_valid_model_input(model):
    if model._interpolator.aesara_function is None:
        raise ValueError('You need to compile graph before. '
                         'See `gempy.set_interpolator`.')
    if model._additional_data.structure_data.df.loc[
        'values', 'len surfaces surface_points'].min() < 1:
        raise ValueError('To compute the model is necessary at least 2 interface '
                         'points per layer')
    if len(model._interpolator.len_series_i) != len(
        model._interpolator.len_series_o):
        raise ValueError('Every Series/Fault need at least 1 orientation and 2 '
                         'surfaces points.')
    is_basement_in_sp = model._surfaces.basement.isin(
        model._surface_points.df['surface']).any()
    is_basement_in_ori = model._surfaces.basement.isin(
        model._orientations.df['surface']).any()

    if is_basement_in_ori or is_basement_in_sp:
        raise ValueError('There are surface points or orientations assigned to the '
                         'Surface defined as basement (bottom of the stack). The '
                         'basement surface only refers to the volume below the last '
                         'surface and is not supposed to be interpolated. '
                         'Add a "basement" surface (`model.add_surface("basement")`)'
                         ' or delete the discordant surface points or orientations')

    last_feature_is_fault = model._stack.df['BottomRelation'].last == 'fault'
    if last_feature_is_fault:
        raise ValueError('Last feature of the stack should not be a fault. '
                         'Reorder the stack using geo_model.reorder_features(List)')


def compute_model_at(new_grid: Union[ndarray], model: Project, **kwargs):
    """This function creates a new custom grid and deactivate all the other
    grids and compute the model there:

    This function does the same as  plus the addition functionality of
        :func:`compute_model`
        passing a given array of points where evaluate the model instead of
        using the :class:`gempy.core.data.GridClass`.

    Args:
        new_grid:
        model (Project):
        kwargs: :func:`compute_model` arguments

    Returns:
        :class:`Solution`
    """
    # #TODO create backup of the mesh and a method to go back to it
    #     set_grid(model, Grid('custom_grid', custom_grid=new_grid))
    warnings.warn('compute_model_at will be deprecated.'
                  'Use argument `at` in compute_model instead', DeprecationWarning)
    model._grid.deactivate_all_grids()
    model.set_custom_grid(new_grid)

    # Now we are good to compute the model again only in the new point
    sol = compute_model(model, set_solutions=False, **kwargs)
    return sol

# endregion


# region activate
@_setdoc_pro([Project.__doc__], )
def activate_interactive_df(geo_model: Project, plot_object=None):
    """Experimental: Activate the use of the QgridProjectIntegration: TODO
    evaluate the use of this functionality

    Notes: Since this feature is for advance levels we will keep only object
    oriented functionality. Should we add in the future,

    TODO: copy docstrings to QgridModelIntegration :param geo_model: [s0]
    :param plot_object: GemPy plot object (so far only vtk is available)

    Args:
        geo_model (Project):
        plot_object:

    Returns:
        :class:`QgridModelIntegration`
    """
    try:
        from gempy_legacy.core.qgrid_integration import QgridModelIntegration
    except ImportError:
        raise ImportError('qgrid package is not installed. No interactive dataframes available.')
    geo_model.qi = QgridModelIntegration(geo_model, plot_object)
    return geo_model.qi
# endregion




from typing import Union

import sys

import numpy as np
import pandas as pn

from .colors import Colors
from ..utils.meta import _setdoc_pro


class Surfaces(object):
    """
    Class that contains the surfaces of the model and the values of each of them.

    Args:
        surface_names (list or np.ndarray): list containing the names of the surfaces
        series (:class:`Series`): [s0]
        values_array (np.ndarray): 2D array with the values of each surface
        properties names (list or np.ndarray): list containing the names of each properties

    Attributes:
        df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the surface names mapped to series and
            the value used for each voxel in the final model.
        series (:class:`Series`)
        colors (:class:`Colors`)

    """

    def __init__(self, series, surface_names=None, values_array=None, properties_names=None):

        self._columns = ['surface', 'series', 'order_surfaces',
                         'isBasement', 'isFault', 'isActive', 'hasData', 'color',
                         'vertices', 'edges', 'sfai', 'id']

        self._columns_vis_drop = ['vertices', 'edges', 'sfai', 'isBasement', 'isFault',
                                  'isActive', 'hasData']
        self._n_properties = len(self._columns) - 1
        self.series = series
        self.colors = Colors(self)

        df_ = pn.DataFrame(columns=self._columns)
        self.df = df_.astype({'surface': str, 'series': 'category',
                              'order_surfaces': int,
                              'isBasement': bool, 'isFault': bool, 'isActive': bool, 'hasData': bool,
                              'color': bool, 'id': int, 'vertices': object, 'edges': object})

        if (np.array(sys.version_info[:2]) <= np.array([3, 6])).all():
            self.df: pn.DataFrame

        self.df['series'] = self.df['series'].cat.add_categories(['Default series'])
        if surface_names is not None:
            self.set_surfaces_names(surface_names)

        if values_array is not None:
            self.set_surfaces_values(values_array=values_array,
                                     properties_names=properties_names)

    def __repr__(self):
        c_ = self.df.columns[~(self.df.columns.isin(self._columns_vis_drop))]

        return self.df[c_].to_string()

    def _repr_html_(self):
        c_ = self.df.columns[~(self.df.columns.isin(self._columns_vis_drop))]

        return self.df[c_].style.applymap(self.background_color, subset=['color']).to_html()

    @property
    def properties_val(self):
        all_col = self.df.columns
        prop_cols = all_col.drop(self._columns)
        return prop_cols.insert(0, 'id')

    @property
    def basement(self):
        return self.df['surface'][self.df['isBasement']]

    def update_id(self, id_list: list = None):
        """
        Set id of the layers (1 based)
        Args:
            id_list (list):

        Returns:
             :class:`Surfaces`:

        """
        self.map_faults()
        if id_list is None:
            # This id is necessary for the faults
            id_unique = self.df.reset_index().index + 1

        self.df['id'] = id_unique

        return self

    def map_faults(self):
        self.df['isFault'] = self.df['series'].map(self.series.faults.df['isFault'])

    @staticmethod
    def background_color(value):
        if isinstance(value, str):
            return "background-color: %s" % value

    # region set formation names
    def set_surfaces_names(self, surfaces_list: list, update_df=True):
        """
         Method to set the names of the surfaces in order. This applies in the surface column of the df

         Args:
             surfaces_list (list[str]):  list of names of surfaces. They are ordered.
             update_df (bool): Update Surfaces.df columns with the default values

         Returns:
             :class:`Surfaces`:
         """
        if isinstance(surfaces_list, (list, np.ndarray)):
            surfaces_list = np.asarray(surfaces_list)

        else:
            raise AttributeError('list_names must be either array_like type')

        # Deleting all columns if they exist
        # TODO check if some of the names are in the df and not deleting them?
        self.df.drop(self.df.index, inplace=True)
        self.df['surface'] = surfaces_list

        # Changing the name of the series is the only way to mutate the series object from surfaces
        if update_df is True:
            self.map_series()
            self.update_id()
            self.set_basement()
            self.reset_order_surfaces()
            self.colors.update_colors()
        return self

    def set_default_surface_name(self):
        """
        Set the minimum number of surfaces to compute a model i.e. surfaces_names: surface1 and basement

        Returns:
             :class:`Surfaces`:

        """
        if self.df.shape[0] == 0:
            # TODO DEBUG: I am not sure that surfaces always has at least one entry. Check it
            self.set_surfaces_names(['surface1', 'basement'])
        return self

    def set_surfaces_names_from_surface_points(self, surface_points):
        """
        Set surfaces names from a :class:`Surface_points` object. This can be useful if the surface points are imported
        from a table.

        Args:
            surface_points (:class:`Surface_points`):

        Returns:

        """
        self.set_surfaces_names(surface_points.df['surface'].unique())
        return self

    def add_surface(self, surface_list: Union[str, list], update_df=True):
        """ Add surface to the df.

        Args:
            surface_list (str, list): name or list of names of the surfaces to apply the functionality
            update_df (bool): Update Surfaces.df columns with the default values

        Returns:
            :class:`gempy.core.data.Surfaces`
        """

        surface_list = np.atleast_1d(surface_list)

        # Remove from the list categories that already exist
        surface_list = surface_list[~np.in1d(surface_list, self.df['surface'].values)]

        for c in surface_list:
            idx = self.df.index.max()
            if idx is np.nan:
                idx = -1
            self.df.loc[idx + 1, 'surface'] = c

        if update_df is True:
            self.map_series()
            self.update_id()
            self.set_basement()
            self.reset_order_surfaces()
            self.colors.update_colors()
        return self

    @_setdoc_pro([update_id.__doc__, pn.DataFrame.drop.__doc__])
    def delete_surface(self, indices: Union[int, str, list, np.ndarray], update_id=True):
        """[s1]

        Args:
            indices (str, list): name or list of names of the series to apply the functionality
            update_id (bool): if true [s0]

        Returns:
             :class:`Surfaces`:

        """
        indices = np.atleast_1d(indices)
        if indices.dtype == int:
            self.df.drop(indices, inplace=True)
        else:
            self.df.drop(self.df.index[self.df['surface'].isin(indices)], inplace=True)

        if update_id is True:
            self.update_id()
            self.set_basement()
            self.reset_order_surfaces()
        return self

    def rename_surfaces(self, to_replace: Union[str, list, dict], **kwargs):
        """Replace values given in to_replace with value.

        Args:
            to_replace (str, regex, list, dict, Series, int, float, or None) –
             How to find the values that will be replaced.
            **kwargs:

        Returns:
            :class:`gempy.core.data.Surfaces`

        See Also:
            :any:`pandas.Series.replace`

        """
        if np.isin(to_replace, self.df['surface']).any():
            print('Two surfaces cannot have the same name.')
        else:
            self.df['surface'].replace(to_replace, inplace=True, **kwargs)
        return self

    def reset_order_surfaces(self):
        self.df['order_surfaces'] = self.df.groupby('series').cumcount() + 1

    def modify_order_surfaces(self, new_value: int, idx: int, series_name: str = None):
        """
        Replace to the new location the old series

        Args:
            new_value (int): New location
            idx (int): Index of the surface to be moved
            series_name (str): name of the series to be moved

        Returns:
            :class:`gempy.core.data.Surfaces`

        """
        if series_name is None:
            series_name = self.df.loc[idx, 'series']

        group = self.df.groupby('series').get_group(series_name)['order_surfaces']
        assert np.isin(new_value, group), 'new_value must exist already in the order_surfaces group.'
        old_value = group[idx]
        self.df.loc[group.index.astype('int'), 'order_surfaces'] = group.replace([new_value, old_value], [old_value, new_value])
        self.sort_surfaces()
        self.set_basement()
        return self

    def sort_surfaces(self):
        """Sort surfaces by series and order_surfaces"""

        self.df.sort_values(by=['series', 'order_surfaces'], inplace=True)
        self.update_id()
        return self.df

    def set_basement(self):
        """
        Set isBasement property to true to the last series of the df.

        Returns:
             :class:`Surfaces`:

        """

        self.df['isBasement'] = False
        idx = self.df.last_valid_index()
        if idx is not None:
            self.df.loc[idx, 'isBasement'] = True

        # TODO add functionality of passing the basement and calling reorder to push basement surface to the bottom
        #  of the data frame
        assert self.df['isBasement'].values.astype(bool).sum() <= 1, 'Only one surface can be basement'
        return self

    # endregion

    # set_series
    def map_series(self, mapping_object: Union[dict, pn.DataFrame] = None):
        """
        Method to map which series every surface belongs to. This step is necessary to assign differenct tectonics
        such as unconformities or faults.


        Args:
            mapping_object (dict, :class:`pn.DataFrame`):
                * dict: keys are the series and values the surfaces belonging to that series

                * pn.DataFrame: Dataframe with surfaces as index and a column series with the corresponding series name
                  of each surface

        Returns:
             :class:`Surfaces`

        """

        # Updating surfaces['series'] categories
        self.df['series'] = self.df['series'].cat.set_categories(self.series.df.index)
        # TODO Fixing this. It is overriding the formations already mapped
        if mapping_object is not None:
            # If none is passed and series exists we will take the name of the first series as a default

            if type(mapping_object) is dict:

                s = []
                f = []
                for k, v in mapping_object.items():
                    for form in np.atleast_1d(v):
                        s.append(k)
                        f.append(form)

                new_series_mapping = pn.DataFrame(list(zip(pn.Categorical(s, self.series.df.index))),
                                                 index=f, columns=['series'])

            elif isinstance(mapping_object, pn.Categorical):
                # This condition is for the case we have surface on the index and in 'series' the category
                # TODO Test this
                new_series_mapping = mapping_object

            else:
                raise AttributeError(str(type(mapping_object)) + ' is not the right attribute type.')

            # Checking which surfaces are on the list to be mapped
            b = self.df['surface'].isin(new_series_mapping.index)
            idx = self.df.index[b]

            # Mapping
            self.df.loc[idx, 'series'] = self.df.loc[idx, 'surface'].map(new_series_mapping['series'])

        # Fill nans
        self.df['series'].fillna(self.series.df.index.values[-1], inplace=True)

        # Reorganize the pile
        self.reset_order_surfaces()
        self.sort_surfaces()
        self.set_basement()

        return self

    # endregion

    # region update_id

    # endregion

    def add_surfaces_values(self, values_array: Union[np.ndarray, list], properties_names: list = np.empty(0)):
        """Add values to be interpolated for each surfaces.

        Args:
            values_array (np.ndarray, list): array-like of the same length as number of surfaces. This functionality
             can be used to assign different geophysical properties to each layer
            properties_names (list): list of names for each values_array columns. This must be of the same size as
             values_array axis 1. By default properties will take the column name: 'value_X'.

        Returns:
            :class:`gempy.core.data.Surfaces`

        """
        values_array = np.atleast_2d(values_array)
        properties_names = np.atleast_1d(properties_names)
        if properties_names.shape[0] != values_array.shape[0]:
            for i in range(values_array.shape[0]):
                properties_names = np.append(properties_names, 'value_' + str(i))

        for e, p_name in enumerate(properties_names):
            try:
                self.df.loc[:, p_name] = values_array[e]
            except ValueError:
                raise ValueError('value_array must have the same length in axis 0 as the number of surfaces')
        return self

    def delete_surface_values(self, properties_names: Union[str, list]):
        """Delete a property or several properties column.

        Args:
            properties_names (str, list[str]): Name of the property to delete

        Returns:
             :class:`gempy.core.data.Surfaces`

        """

        properties_names = np.asarray(properties_names)
        self.df.drop(properties_names, axis=1, inplace=True)
        return True

    def set_surfaces_values(self, values_array: Union[np.ndarray, list], properties_names: list = np.empty(0)):
        """Set values to be interpolated for each surfaces. This method will delete the previous values.

        Args:
            values_array (np.ndarray, list): array-like of the same length as number of surfaces. This functionality
             can be used to assign different geophysical properties to each layer
            properties_names (list): list of names for each values_array columns. This must be of same size as
             values_array axis 1. By default properties will take the column name: 'value_X'.

        Returns:
             :class:`gempy.core.data.Surfaces`

        """
        # Check if there are values columns already
        old_prop_names = self.df.columns[~self.df.columns.isin(['surface', 'series', 'order_surfaces',
                                                                'id', 'isBasement', 'color'])]
        # Delete old
        self.delete_surface_values(old_prop_names)

        # Create new
        self.add_surfaces_values(values_array, properties_names)
        return self

    def modify_surface_values(self, idx, properties_names, values):
        """Method to modify values using loc of pandas.

        Args:
            idx (int, list[int]):
            properties_names (str, list[str]):
            values (float, np.ndarray):

        Returns:
             :class:`gempy.core.data.Surfaces`

        """
        properties_names = np.atleast_1d(properties_names)
        assert ~np.isin(properties_names, ['surface', 'series', 'order_surfaces', 'id', 'isBasement', 'color']), \
            'only property names can be modified with this method'

        self.df.loc[idx, properties_names] = values
        return self




from matplotlib.cm import ScalarMappable as SM
from gempy_legacy.plot._visualization_2d import PlotData2D
import numpy as np
import os
import itertools as it


def export_geomap2geotiff(path, geo_model, geo_map=None, geotiff_filepath=None):
    """

    Args:
        path (str): Filepath for the exported geotiff, must end in .tif
        geo_map (np.ndarray): 2-D array containing the geological map
        cmap (matplotlib colormap): The colormap to be used for the export
        geotiff_filepath (str): Filepath of the template geotiff

    Returns:
        Saves the geological map as a geotiff to the given path.
    """
    from osgeo import gdal

    plot = PlotData2D(geo_model)
    cmap = plot._cmap
    norm = plot._norm

    if geo_map is None:
        geo_map = geo_model.solutions.geological_map[0].reshape(geo_model._grid.topography.resolution)

    if geotiff_filepath is None:
        # call the other function
        print('stupid')

    # **********************************************************************
    geo_map_rgb = SM(norm=norm, cmap=cmap).to_rgba(geo_map.T) # r,g,b,alpha
    # **********************************************************************
    # gdal.UseExceptions()
    ds = gdal.Open(geotiff_filepath)
    band = ds.GetRasterBand(1)
    arr = band.ReadAsArray()
    [cols, rows] = arr.shape

    outFileName = path
    driver = gdal.GetDriverByName("GTiff")
    options = ['PROFILE=GeoTiff', 'PHOTOMETRIC=RGB', 'COMPRESS=JPEG']
    outdata = driver.Create(outFileName, rows, cols, 3, gdal.GDT_Byte, options=options)

    outdata.SetGeoTransform(ds.GetGeoTransform())  # sets same geotransform as input
    outdata.SetProjection(ds.GetProjection())  # sets same projection as input
    outdata.GetRasterBand(1).WriteArray(geo_map_rgb[:, ::-1, 0].T * 256)
    outdata.GetRasterBand(2).WriteArray(geo_map_rgb[:, ::-1, 1].T * 256)
    outdata.GetRasterBand(3).WriteArray(geo_map_rgb[:, ::-1, 2].T * 256)
    outdata.GetRasterBand(1).SetColorInterpretation(gdal.GCI_RedBand)
    outdata.GetRasterBand(2).SetColorInterpretation(gdal.GCI_GreenBand)
    outdata.GetRasterBand(3).SetColorInterpretation(gdal.GCI_BlueBand)
    # outdata.GetRasterBand(4).SetColorInterpretation(gdal.GCI_AlphaBand)  # alpha band

    # outdata.GetRasterBand(1).SetNoDataValue(999)##if you want these values transparent
    outdata.FlushCache()  # saves to disk
    outdata = None  # closes file (important)
    band = None
    ds = None

    print("Successfully exported geological map to  " +path)

def export_moose_input(geo_model, path=None, filename='geo_model_units_moose_input.i'):
    """
    Method to export a 3D geological model as MOOSE compatible input. 

    Args:
        path (str): Filepath for the exported input file
        filename (str): name of exported input file

    Returns:
        
    """
    # get model dimensions
    nx, ny, nz = geo_model.grid.regular_grid.resolution
    xmin, xmax, ymin, ymax, zmin, zmax = geo_model.solutions.grid.regular_grid.extent
    
    # get unit IDs and restructure them
    ids = np.round(geo_model.solutions.lith_block)
    ids = ids.astype(int)
    
    liths = ids.reshape((nx, ny, nz))
    liths = liths.flatten('F')

    # create unit ID string for the fstring
    idstring = '\n  '.join(map(str, liths))

    # create a dictionary with unit names and corresponding unit IDs
    sids = dict(zip(geo_model._surfaces.df['surface'], geo_model._surfaces.df['id']))
    surfs = list(sids.keys())
    uids = list(sids.values())
    # create strings for fstring, so in MOOSE, units have a name instead of an ID
    surfs_string = ' '.join(surfs)
    ids_string = ' '.join(map(str, uids))
    
    fstring = f"""[MeshGenerators]
  [./gmg]
  type = GeneratedMeshGenerator
  dim = 3
  nx = {nx}
  ny = {ny}
  nz = {nz}
  xmin = {xmin}
  xmax = {xmax}
  yim = {ymin}
  ymax = {ymax}
  zmin = {zmin}
  zmax = {zmax}
  block_id = '{ids_string}'
  block_name = '{surfs_string}'
  [../]
  
  [./subdomains]
    type = ElementSubdomainIDGenerator
    input = gmg
    subdomain_ids = '{idstring}'
  [../]
[]

[Mesh]
  type = MeshGeneratorMesh
[]
"""
    if not path:
        path = './'
    if not os.path.exists(path):
      os.makedirs(path)

    f = open(path+filename, 'w+')
    
    f.write(fstring)
    f.close()
    
    print("Successfully exported geological model as moose input to "+path)

def export_shemat_suite_input_file(geo_model, path: str=None, filename: str='geo_model_SHEMAT_input'):
    """
    Method to export a 3D geological model as SHEMAT-Suite input-file for a conductive HT-simulation. 

    Args:
        path (str): Filepath for the exported input file (default './')
        filename (str): name of exported input file (default 'geo_model_SHEMAT_input')
    """
    # get model dimensions
    nx, ny, nz = geo_model.grid.regular_grid.resolution
    xmin, xmax, ymin, ymax, zmin, zmax = geo_model.solutions.grid.regular_grid.extent
    
    delx = (xmax - xmin)/nx
    dely = (ymax - ymin)/ny
    delz = (zmax - zmin)/nz
    
    # get unit IDs and restructure them
    ids = np.round(geo_model.solutions.lith_block)
    ids = ids.astype(int)
    
    liths = ids.reshape((nx, ny, nz))
    liths = liths.flatten('F')

    # group litho in space-saving way
    sequence = [len(list(group)) for key, group in it.groupby(liths)]
    unit_id = [key for key, group in it.groupby(liths)]
    combined = ["%s*%s" % (pair) for pair in zip(sequence,unit_id)]

    combined_string = " ".join(combined)

    # get number of units and set units string
    units = geo_model.surfaces.df[['surface', 'id']]
    
    unitstring = ""
    for index, rows in units.iterrows():
        unitstring += f"0.01d-10    1.d0  1.d0  1.e-14	 1.e-10  1.d0  1.d0  3.74	0.  2077074.  10  2e-3	!{rows['surface']} \n" 	

    # input file as f-string
    fstring = f"""!==========>>>>> INFO
# Title
{filename}

# linfo
1 2 1 1

# runmode
1

# timestep control
0
1           1           0           0

# tunit
1
 
# time periods, records=1
0      60000000    200      lin
           
# output times, records=10
1
6000000
12000000
18000000
24000000
30000000
36000000
42000000
48000000
54000000
    
# file output: hdf vtk

# active temp

# PROPS=bas

# USER=none


# grid
{nx} {ny} {nz}

# delx
{nx}*{delx}

# dely
{ny}*{dely}

# delz
{nz}*{delz}

!==========>>>>> NONLINEAR SOLVER
# nlsolve
50 0

!==========>>>>> FLOW
# lsolvef (linear solver control)
1.d-12 64 500
# nliterf (nonlinear iteration control)
1.0d-10 1.0

!==========>>>>> TEMPERATURE
# lsolvet (linear solver control)
1.d-12 64 500
# nlitert (nonlinear iteration control)
1.0d-10 1.0

!==========>>>>> INITIAL VALUES
# temp init
{nx*ny*nz}*15.0d0  

# head init
{nx*ny*nz}*7500

!==========>>>>> UNIT DESCRIPTION
!!
# units
{unitstring}

!==========>>>>>   define boundary properties
# temp bcd, simple=top, value=init

# temp bcn, simple=base, error=ignore
{nx*ny}*0.06

# uindex
{combined_string}"""

    if not path:
        path = './'
    if not os.path.exists(path):
        os.makedirs(path)

    f = open(path+filename, 'w+')
    
    f.write(fstring)
    f.close()
    
    print("Successfully exported geological model as SHEMAT-Suite input to "+path) 


def export_pflotran_input(geo_model, path=None, filename='pflotran.ugi'):
    """
    Method to export a 3D geological model as PFLOTRAN implicit unstructured grid
    (see pflotran.org)

    Args:
        path (str): Filepath for the exported input file
        filename (str): name of exported input file

    Returns:
        
    """
    #
    # Added by Moise Rousseau, December 8th, 2020
    # see https://www.pflotran.org/documentation/user_guide/cards/subsurface/grid_card.html
    #
    # create vertices, elements and groups
    vertices, elements, groups = __build_vertices_elements_groups__(geo_model)
    
    #create mesh
    if not path:
        path = './'
    if not os.path.exists(path):
        os.makedirs(path)
    ext = filename.split('.')[-1]
    if ext == 'ugi': 
        mesh_format = 'ascii'
    elif ext == 'h5': 
        mesh_format = 'hdf5'
        try:
            import h5py
        except:
            print("h5py library not installed. Please install it to enable hdf5 output")
            print("Export as ascii instead")
            mesh_format = 'ascii'
            filename += '.ugi'
    else: #assume ugi mesh
        mesh_format = 'ascii'
        filename += '.ugi'
    
    #export mesh
    if mesh_format == 'ascii': #export as ugi
        out = open(path+filename, 'w')
        out.write(f"{len(elements)} {len(vertices)}")
        for element in elements:
            out.write('\nH')
            for x in element[1:]:
              out.write(f" {int(x)}")
        for vertice in vertices:
            out.write('\n')
            out.write(f"{vertice[0]} {vertice[1]} {vertice[2]}")
        out.close()
    elif mesh_format == 'hdf5':
        out = h5py.File(path+filename, 'w')
        out.create_dataset("Domain/Cells", data=elements)
        out.create_dataset("Domain/Vertices", data=vertices)
        out.close()
    
    #topography z faces
    nx, ny, nz = geo_model.grid.regular_grid.resolution
    if len(elements) != nx*ny*nz:
        inter = __generate_topographic_surface__(vertices,elements)
        if mesh_format == 'ascii':
            out = open(path+"topography_surface.ss",'w')
            out.write(str(len(inter))+'\n')
            for elem in inter:
                out.write("Q ")
                for node in elem[1:]:
                    out.write(f"{node} ")
                out.write('\n')
            out.close()
        if mesh_format == "hdf5":
            out = h5py.File(path+filename,'r+')
            out.create_dataset(f"Regions/Topography_surface/Vertex Ids", data=inter)
            out.close()
    
    #export groups
    for grp_name,cell_ids in groups.items():
        if mesh_format == 'ascii':
            out = open(path+grp_name+'.vs','w')
            for x in cell_ids:
                out.write(f"{x}\n")
            out.close()
        if mesh_format == "hdf5":
            out = h5py.File(path+filename,'r+')
            out.create_dataset(f"Regions/{grp_name}/Cell Ids", data=cell_ids)
            out.close()
            
    print("Successfully exported geological model as PFLOTRAN input to "+path+filename)
    return

def export_flac3D_input(geo_model, path=None, filename='geomodel.f3grid'):
    """
    Method to export a 3D geological model FLAC3D readable format

    Args:
        path (str): Filepath for the exported input file
        filename (str): name of exported input file

    Returns:
        
    """
    print("Warning! This is experimental and this have not been validated")
    #
    # Added by Moise Rousseau, December 9th, 2020
    #
    # create vertices and elements
    vertices, elements, groups = __build_vertices_elements_groups__(geo_model)
    
    #open output file
    #if not path:
    #    path = './'
    #if not os.path.exists(path):
    #    os.makedirs(path)

    out = open(path+filename, 'w')
    
    #write gridpoints
    out.write("*GRIDPOINTS")
    for i,vertice in enumerate(vertices):
        out.write(f"\nG {i+1} {vertice[0]} {vertice[1]} {vertice[2]}")
    
    #write elements
    out.write('\n*ZONES')
    for i,elem in enumerate(elements):
        out.write('\nB8')
        for x in elem[1:]:
            out.write(f" {x}")
    
    #make groups
    out.write('\n*GROUPS\n')
    for grp_name,grp in groups.items():
        out.write(f"*ZGROUP {grp_name}\n")
        count = 0
        for x in grp:
            out.write(f"{x} ")
            count += 1
            if count == 8:
                out.write("\n")
                count = 0
        if count != 0: out.write("\n")
    
    out.close()
    print("Successfully exported geological model as FLAC3D input to "+path)
    return



def __build_vertices_elements_groups__(geo_model):
    """
    In: GemPy model class instance
        Generate topo interface (optional)
    Out: Vertices coordinates numpy array
         Elements defined by the number of vertices (always 8)
           and their vertices id (bottom face and top face)
    """
    # get model information
    nx, ny, nz = geo_model.grid.regular_grid.resolution
    xmin, xmax, ymin, ymax, zmin, zmax = geo_model.solutions.grid.regular_grid.extent
    
    # create vertices array
    dx, dy, dz = (xmax-xmin)/nx, (ymax-ymin)/ny, (zmax-zmin)/nz
    n_vertices = (nx+1) * (ny+1) * (nz+1)
    vertices = np.zeros((n_vertices, 3), dtype='f8')
    vertices_ids = np.arange(n_vertices) #used to generate coordinate
    vertices[:,0] = vertices_ids % (nx+1) * dx + xmin
    vertices[:,1] = ( vertices_ids % ( (nx+1)*(ny+1) ) ) // (nx+1) * dy + ymin
    vertices[:,2] = vertices_ids // ( (nx+1)*(ny+1) ) * dz + zmin
    
    #build elements
    n_elements = nx*ny*nz
    element_ids = np.arange(n_elements) #used to generate elems
    elements = np.zeros((n_elements,9), dtype='i8')
    i = element_ids % nz
    j = element_ids // nz % ny
    k = element_ids // (nz*ny)
    elements[:,0] = 8 #all hex
    elements[:,1] = 1 + i*(nx+1)*(ny+1) + j*(nx+1) + k
    elements[:,2] = elements[:,1] + 1
    elements[:,3] = elements[:,2] + (nx+1)
    elements[:,4] = elements[:,3] - 1
    elements[:,5] = elements[:,1] + ( (nx+1)*(ny+1) )
    elements[:,6] = elements[:,5] + 1
    elements[:,7] = elements[:,6] + (nx+1)
    elements[:,8] = elements[:,7] - 1
    
    #build groups
    lith_ids = np.round(geo_model.solutions.lith_block)
    lith_ids = lith_ids.astype(int)
    sids = dict(zip(geo_model._surfaces.df['surface'], geo_model._surfaces.df['id']))
    groups = {}
    for region_name,region_id in sids.items():
        cell_ids = np.where(lith_ids == region_id)[0] + 1
        if not len(cell_ids): continue
        groups[region_name] = cell_ids
            
    #remove element above topography
    lith_ids = np.round(geo_model.solutions.lith_block)
    lith_ids = lith_ids.astype(int)
    inactive_cells = ~np.isin(lith_ids, geo_model._surfaces.df['id']) #get cell not on a lith_block
    if np.any(inactive_cells):
        #update correspondance
        new_id_vertices = np.zeros(len(vertices),dtype='i8')
        new_id_elements = np.zeros(len(elements),dtype='i8')
        #remove inactive cell
        elements = elements[~inactive_cells]
        new_id_elements[~inactive_cells] = np.arange(len(elements))+1
        #remove deleted vertices
        cond = np.isin(np.arange(len(vertices))+1, elements[:,1:].flatten())
        vertices = vertices[cond]
        new_id_vertices[cond] = np.arange(len(vertices))+1
        #renumber vertices in element
        elements[1:] = new_id_vertices[elements[1:]-1]
        #renumber groups
        for grp in groups.values():
          grp[:] = new_id_elements[grp-1]
    
    return vertices, elements, groups


def __generate_topographic_surface__(vertices, elements):
    """
    In: Vertices array generated by __build_vertices_and_elements__
        Elements array generated by __build_vertices_and_elements__
    Out: Topography surface defined by the top faces
    """
    inter = np.zeros((len(elements),5), dtype='i8')
    count = 1
    inter[:,0] = 4 #all quad face
    inter[0, 1:] = elements[-1,5:] 
    last_node_z = -1e20
    for ie,elem in enumerate(elements):
        node = elem[-1]
        node_z = vertices[node-1][2]
        if node_z < last_node_z:
            inter[count,1:] = elements[ie-1,5:]
            count += 1
        last_node_z = node_z
    inter = inter[:count]
    return inter
        



import gempy_legacy as gp
import numpy as np
import pandas as pd
import pytest
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import os

# Input files
from gempy_legacy.addons.map2gempy import loop2gempy
from gempy_legacy.core.grid_modules.create_topography import GDAL_IMPORT

root = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/data/input_data/turner_syncline/'
root2 = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/data/input_data/australia/'
path = os.path.dirname(__file__) + '/../input_data/'

orientations_file = root + 'orientations_clean.csv'
orientations_file2 = root2 + 'orientations_clean.csv'

contacts_file = root + 'contacts_clean.csv'
contacts_file2 = root2 + 'contacts_clean.csv'
faults_contact = root + 'faults.csv'
faults_contact2 = root2 + 'faults.csv'
faults_orientations = root + 'fault_orientations.csv'
faults_orientations2 = root2 + 'fault_orientations.csv'
series_file = root + 'all_sorts_clean.csv'
series_file2 = root2 + 'all_sorts_clean.csv'

faults_rel_matrix = root + 'fault-fault-relationships.csv'
faults_rel_matrix2 = root2 + 'fault-fault-relationships.csv'
series_rel_matrix = root + 'group-fault-relationships.csv'
series_rel_matrix2 = root2 + 'group-fault-relationships.csv'

ff = root2 + 'fault-fault-relationships.csv'
fg = root2 + 'group-fault-relationships.csv'

fp = path + 'dtm_rp.tif'
fp2 = path + 'dtm_rp2.tif'

bbox = (500000, 7490000, 545000, 7520000)
model_base = -0  # Original 3200
model_top = 800
extent_g = [515687.3100586407, 562666.8601065436,
            7473446.765934078, 7521273.574077863,
            -3200, 1200.0]

extent = [515687.3100586407, 7473446.765934078,
          562666.8601065436, 7521273.574077863,
          -3200, 1200.0]


@pytest.mark.skipif(("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true") or
                    ("GITHUBACTION" in os.environ and os.environ["GITHUBACTION"] == "true"),
                    reason="Skipping this test on Travis CI. For some reason there is a linalg "
                           "error.")
def test_loop2gempy():
    topo = fp
    topo = None

    loop2gempy(contacts_file, orientations_file, bbox, series_file, model_base,
               model_top, topo, faults_contact, faults_orientations, 'testing_map',
               vtk=True, vtk_path=None, image_2d=True)


def test_map2loop2relmatrix():
    ff_ = pd.read_csv(ff).set_index('fault_id')
    fg_ = pd.read_csv(fg).set_index('group')
    p = pd.concat((ff_, fg_.T), axis=1)
    print(p)


@pytest.mark.skipif(("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true") or
                    ("GITHUBACTION" in os.environ and os.environ["GITHUBACTION"] == "true" or
                     GDAL_IMPORT is False),
                    reason="Skipping this test on Travis CI. For some reason there is a linalg "
                           "error.")
def test_loop2gempy2():
    topo = fp2
    # topo = None

    loop2gempy(contacts_file2, orientations_file2, extent[:4], series_file2,
               extent[4],
               extent[5],
               dtm_reproj_file=topo,
               faults_contact=faults_contact2,
               faults_orientations=faults_orientations2,
               faults_faults_rel=ff,
               faults_groups_rel=fg,
               model_name='testing_map',
               compute=True,
               vtk=True, vtk_path=None, image_2d=True)


@pytest.mark.skipif(("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true") or
                    ("GITHUBACTION" in os.environ and os.environ["GITHUBACTION"] == "true" or 
                     GDAL_IMPORT is False
                     ),
                    reason="Skipping this test on Travis CI. For some reason there is a linalg "
                           "error.")
def test_map2loop_model_import_data():
    geo_model = gp.create_model('test_map2Loop')
    gp.init_data(
        geo_model,
        extent=[bbox[0], bbox[2], bbox[1], bbox[3], model_base, model_top],
        resolution=[50, 50, 50],
        path_o=orientations_file,
        path_i=contacts_file
    )

    # Load Topology
    geo_model.set_topography(source='gdal', filepath=fp)

    gp.plot_2d(geo_model, ve=10, show_topography=True)
    plt.show()

    # Plot in 3D
    gp.plot_3d(geo_model, ve=None, show_topography=False, image=True,
               kwargs_plot_data={'arrow_size': 400}
               )
    print(geo_model.orientations)

# skip test i
@pytest.mark.skipif(("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true") or
                    ("GITHUBACTION" in os.environ and os.environ["GITHUBACTION"] == "true" or
                     GDAL_IMPORT is False
                     ),
                    reason="Skipping this test on Travis CI. For some reason there is a linalg "
                           "error.")
def test_map2loop_model_no_faults():
    # Location box
    bbox = (500000, 7490000, 545000, 7520000)
    model_base = -3200  # Original 3200
    model_top = 800

    # Input files
    geo_model = gp.create_model('test_map2Loop')
    gp.init_data(
        geo_model,
        extent=[bbox[0], bbox[2], bbox[1], bbox[3], model_base, model_top],
        resolution=[50, 50, 80],
        path_o=orientations_file,
        path_i=contacts_file
    )

    gp.set_interpolator(geo_model)

    # Load Topology
    geo_model.set_topography(source='gdal', filepath=fp)
    # Stack Processing
    contents = np.genfromtxt(series_file,
                             delimiter=',', dtype='U100')[1:, 4:-1]

    map_series_to_surfaces = {}
    for pair in contents:
        map_series_to_surfaces.setdefault(pair[1], []).append(pair[0])

    gp.map_stack_to_surfaces(geo_model, map_series_to_surfaces,
                             remove_unused_series=False)

    gp.plot_2d(geo_model, ve=10, show_topography=False)
    plt.show()

    # Plot in 3D
    gp.plot_3d(geo_model, ve=10, show_topography=False, image=True)

    # Stack Processing
    contents = np.genfromtxt(series_file,
                             delimiter=',', dtype='U100')[1:, 4:-1]

    map_series_to_surfaces = {}
    for pair in contents:
        map_series_to_surfaces.setdefault(pair[1], []).append(pair[0])

    gp.map_stack_to_surfaces(geo_model, map_series_to_surfaces,
                             remove_unused_series=False)

    # Adding axial rescale
    # geo_model._rescaling.toggle_axial_anisotropy()

    # Increasing nugget effect
    geo_model.modify_surface_points(
        geo_model.surface_points.df.index,
        smooth=0.001
    )

    geo_model.modify_kriging_parameters('drift equations', [9, 9, 9, 9, 9])

    gp.compute_model(geo_model)

    gp.plot_2d(geo_model,
               section_names=['topography'],
               show_topography=True,
               )
    plt.show()

    gp.plot_3d(geo_model, ve=10, show_topography=True,
               image=True,
               show_lith=False,
               )
